#  Copyright (c) 2021 Mireia Bellot
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test module for institution Learner data management """
import logging

import pytest

import tests.utils

from tests.utils import getting_variables


@pytest.mark.django_db
def test_api_institution_new_learner(rest_api_client, institution_course_test_case):
    """

    INSTITUTION TEST: LEARNER
       1) List learners
       2) Create a new learner
       3) Read new learner information
       4) Update new learner
       5) Delete new learner
    """
    # Institution users need mail domain validation
    institution_course_test_case['institution'].mail_domain = 'tesla-ce.eu'
    institution_course_test_case['institution'].save()
    institution_user = institution_course_test_case['user'].institutionuser
    institution_id = institution_course_test_case['institution'].id

    logging.info('\nTESTING INSTITUTIONS: New Learner *********************************************')

    # Get the list of Institutions
    """ ---------------------------------------------------------------------
    LIST LEARNERS:
       GET /api/v2/institution/(int: institution_id)/learner/
       Status Codes:
           200 OK – Ok
           404 Not Found – Institution not found
    Request Headers: Authorization - JWT with Institution Admin (or any Admin: Data Admin, Legal Admin, 
    Global Admin, ,Send Admin) privileges
    """
    logging.info('\n1) LIST LEARNERS --------------------------------------')
    institution_user.inst_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_path = '/api/v2/institution/{}/learner/'.format(institution_id)
    str_message = 'RESPONSE institution_id:{}'.format(institution_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List learners', str_message, 200)
    n_learners = body['count']
    variables = getting_variables(body, institution_id)

    if variables['n_institution'] > 0:
        tests.utils.check_pagination(rest_api_client, body)
        args = [['List Learners'], ['ID FIRST LEARNER: ', variables['id_first_institution']],
                ['NUMBER OF LEARNERS:', variables['n_institution']],
                ['INSTITUTION ID TEST', institution_id],
                ['ID NON EXISTING INSTITUTION:', variables['id_non_existing_institution']]]
    else:
        args = [['List Institutions'], ['Empty Institutions list. Reading first Institution test will be skipped.']]

    tests.utils.print_log(args)

    # 2) Create a new learner
    """ ---------------------------------------------------------------------
    CREATE A NEW LEARNER:
        POST /api/v2/institution/(int:institution_id)/learner/
        Request JSON Object
            uid (string) – Unique ID of this learner in the institution.
            email (string) – Email of the learner. If institution mail_domain is provided, 
                             this email must be in this domain.
            first_name (string) – First name of the learner.
            last_name (string) – Last name of the learner.
        Status Codes:
            201 Created – Created
            400 Bad Request – Invalid information provided. The response contains the description of the errors.
            404 Not Found – Institution not found
    Request Headers: Authorization - JWT with Institution Admin privileges
    """
    institution_user.inst_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    logging.info('\n2) CREATE A NEW LEARNER --------------------------------------')
    str_data = {'uid': 'LEARNER_UID', 'email': 'mail_test@tesla-ce.eu',
                'first_name': 'TEST_LEARNER_NAME',
                'last_name': 'TEST_LEARNER_LASTNAME'}
    new_learner_id = tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                                      'Create a new learner', 'RESPONSE: ', 201)

    # Ensure number of learners has increased
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List learners', 'RESPONSE:', 200)
    assert n_learners + 1 == body['count']

    #  Status 400: Validating email domain
    str_data = {'uid': 'LEARNER_UID_2', 'email': 'mail_test@mail.com',
                'first_name': 'TEST_LEARNER_NAME',
                'last_name': 'TEST_LEARNER_LASTNAME'}
    tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                     'Create a new learner', 'RESPONSE: ', 400)

    # Status 400: Validating unique email
    tests.utils.post_rest_api_client(rest_api_client, str_path,
                                     str_data,
                                     'Create a new learner failed',
                                     'RESPONSE: ', 400)

    # 3) Read learner information
    """ ---------------------------------------------------------------------
    READ LEARNER INFORMATION:
    GET /api/v2/institution/(int: institution_id)/learner/(int: learner_id)/
    Status Codes:
        200 OK – Ok
        404 Not Found – Institution not found
        404 Not Found – Learner not found
    Privileges: Institution Admin (or any Admin) privileges
    """
    logging.info('\n3) READ LEARNER INFORMATION --------------------------------------')
    institution_user.inst_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_path = '/api/v2/institution/{}/learner/{}/'.format(institution_id, new_learner_id)
    str_response = 'RESPONSE Learner ID={}:'.format(new_learner_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'Read Learner Information', str_response, 200)
    # No informed consent for a new learner
    assert body['consent'] is None
    assert body['consent_rejected'] is None
    # No SEND category for a new learner
    assert body['send']['is_send'] is False

    # TODO? Read Learner information errors

    # 4) Update learner
    """ ---------------------------------------------------------------------
    UPDATE LEARNER:
        PUT /api/v2/institution/(int: institution_id)/learner/(int: learner_id)/
        Status Codes:
            200 OK – Ok
            400 Bad Request – Invalid information provided. The response contains the description of the errors.
            404 Not Found – Institution not found
            404 Not Found – Learner not found
    Request Headers: Authorization - JWT with Institution Admin privileges
    """
    logging.info('\n4) UPDATE LEARNER --------------------------------------')
    institution_user.inst_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_data = {'uid': 'LEARNER_UID_UPDATED', 'email': 'mail_test_UPDATED@tesla-ce.eu',
                'first_name': 'TEST_LEARNER_NAME_UPDATED',
                'last_name': 'TEST_LEARNER_LASTNAME_UPDATED'}
    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data,
                                    'Update Learner', 'RESPONSE: ', 200)

    # TODO? Update Learner errors

    # 5) Delete learner
    """ ---------------------------------------------------------------------
    DELETE LEARNER:
        DELETE /api/v2/institution/(int: institution_id)/learner/(int: learner_id)/
        Status Codes:
            200 OK – Ok
            400 Bad Request – Invalid information provided. The response contains the description of the errors.
            404 Not Found – Institution not found
            404 Not Found – Learner not found
    Request Headers: Authorization - JWT with Institution Admin/Data privileges
    """
    institution_user.inst_admin = True
    # 666 TODO: Add authorization for deleting learner by an Institution Data admin user?
    # institution_user.data_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    logging.info('\n5) DELETE LEARNER --------------------------------------')

    str_path = '/api/v2/institution/{}/learner/{}/'.format(institution_id, new_learner_id)
    tests.utils.delete_rest_api_client(rest_api_client, str_path,
                                       'Delete Learner', "RESPONSE: ", 204)

    # Ensure number of learners has decreased
    str_path = '/api/v2/institution/{}/learner/'.format(institution_id)
    str_message = 'RESPONSE institution_id:{}'.format(institution_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List learners', str_message, 200)
    assert body['count'] == n_learners


@pytest.mark.django_db
def test_api_institution_learner_ic(rest_api_client, institution_course_test_case):
    """

    INSTITUTION TEST: LEARNER'S INFORMED CONSENT
       1) Accept an informed consent for a learner
       2) Reject current Informed Consent of a learner
    """
    institution_user = institution_course_test_case['user'].institutionuser
    institution_id = institution_course_test_case['institution'].id
    learner = institution_course_test_case['learner']
    learner_user = learner.institutionuser
    learner_user_id = learner_user.id

    logging.info('\nTESTING INSTITUTIONS: Learner\'s Informed Consent *********************************************')

    # 1) Accept an informed consent for a learner
    """ ---------------------------------------------------------------------
    ACCEPT INFORMED CONSENT FOR A LEARNER:
        POST /api/v2/institution/(int: institution_id)/learner/(int: learner_id)/ic/
        Request JSON Object: version (string) – Informed consent version to assign
        Status Codes:
            200 OK – Ok
            400 Bad Request – Invalid information provided. The response contains the description of the errors.
            400 Not Found – Institution not found
            404 Not Found – Learner not found
    Request Headers: Authorization - JWT with LEARNER privileges (only own learner)
    """
    logging.info('\n1) ACCEPT INFORMED CONSENT FOR A LEARNER --------------------------------------')
    rest_api_client.force_authenticate(user=learner_user)

    # Getting valid informed consent for current institution
    ic_str_path = '/api/v2/institution/{}/ic/'.format(learner_user_id)
    ic_str_response = 'RESPONSE institution_id={}:'.format(institution_id)
    ic_body = tests.utils.get_rest_api_client(rest_api_client, ic_str_path,
                                              'List Informed Consents', ic_str_response, 200)

    # If no valid informed consent exist: create one
    if ic_body['count'] > 0:
        ic_version = ic_body['results'][0].version
        str_path = '/api/v2/institution/{}/learner/{}/ic'.format(institution_id, learner_user_id)
        str_data = {'version': ic_version}
        tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                         'Accept Informed Consent for a Learner',
                                         'RESPONSE: ', 200)
    # If valid informed consent exist
    else:
        str_path = '/api/v2/institution/{}/ic/'.format(institution_id)
        str_data = {'version': '0.0.1', 'valid_from': '2021-06-01T10:00'}
        tests.utils.post_rest_api_client(rest_api_client, str_path,
                                         str_data, 'Create a new Informed Consent',
                                         'RESPONSE:', 201)
        rest_api_client.force_authenticate(user=learner_user)
        str_path = '/api/v2/institution/{}/learner/{}/ic/'.format(institution_id, learner_user_id)
        str_data = {'version': '0.0.1'}
        tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                         'Accept Informed Consent for a Learner',
                                         'RESPONSE: ', 200)

    institution_user.inst_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)
    str_path = '/api/v2/institution/{}/learner/{}/'.format(institution_id, learner_user_id)
    str_response = 'RESPONSE Learner ID={}:'.format(learner_user_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'Read Learner Information', str_response, 200)
    assert body['consent'] is not None

    # 2) Reject current Informed Consent of a learner
    """ ---------------------------------------------------------------------
    REJECT CURRENT INFORMED CONSENT OF A LEARNER:
        DELETE /api/v2/institution/(int: institution_id)/learner/(int: learner_id)/ic/
        Status Codes:
            200 OK – Ok
            400 Bad Request – Invalid information provided. The response contains the description of the errors.
            400 Not Found – Institution not found
            404 Not Found – Learner not found
    Request Headers: Authorization - JWT with LEARNER privileges
    """
    logging.info('\n2) REJECT INFORMED CONSENT FOR A LEARNER --------------------------------------')
    rest_api_client.force_authenticate(user=learner_user)
    str_path = '/api/v2/institution/{}/learner/{}/ic/'.format(institution_id, learner_user_id)
    tests.utils.delete_rest_api_client(rest_api_client, str_path,
                                       'Reject Current Informed Consent of a Learner', "RESPONSE: ", 200)

    rest_api_client.force_authenticate(user=institution_user)
    str_path = '/api/v2/institution/{}/learner/{}/'.format(institution_id, learner_user_id)
    str_response = 'RESPONSE Learner ID={}:'.format(learner_user_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'Read Learner Information', str_response, 200)
    assert body['consent_rejected'] is not None


@pytest.mark.django_db
def test_api_institution_learner_send(rest_api_client, institution_course_test_case):
    """

    INSTITUTION TEST: LEARNER'S SEND CATEGORY
       1) Read SEND categories assigned to a learner
       2) Add SEND category to a learner
       3) Remove a SEND Category from a learner
    """
    institution_user = institution_course_test_case['user'].institutionuser
    institution_id = institution_course_test_case['institution'].id
    learner = institution_course_test_case['learner']
    learner_user = learner.institutionuser
    learner_user_id = learner_user.id

    logging.info('\nTESTING INSTITUTIONS: Learner\'s SEND category *********************************************')

    # 1) Read SEND categories assigned to a learner
    """ ---------------------------------------------------------------------
    Read SEND categories assigned to a learner:
        GET /api/v2/institution/(int: institution_id)/learner/(int: learner_id)/send/
        Status Codes:
            200 OK – Ok
            400 Bad Request – Invalid information provided. The response contains the description of the errors.
            404 Not Found – Institution not found
            404 Not Found – Learner not found
    Request Headers: Authorization - JWT with Institution Admin/SEND privileges
    """
    logging.info('\n1) Read SEND categories assigned to a learner --------------------------------------')
    institution_user.inst_admin = False
    institution_user.send_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_path = '/api/v2/institution/{}/learner/{}/send/'.format(institution_id, learner_user_id)
    str_response = 'RESPONSE Learner ID={}:'.format(learner_user_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'Read SEND categories assigned to a learner',
                                           str_response, 200)
    # No SEND category for a new learner
    assert body['count'] == 0

    # 2) Add SEND category to a learner
    """ ---------------------------------------------------------------------
    Add SEND category to a learner:
        POST /api/v2/institution/(int: institution_id)/learner/(int: learner_id)/send/
        Request JSON Object:
            category (int) – SEND Category ID
            expires_at (datetime) – When the special need is temporal, provide the date when it disappears. 
            For permanent needs, let it null.
        Status Codes:
            201 Created – Created
            400 Bad Request – Invalid information provided. The response contains the description of the errors.
            404 Not Found – Institution not found
            404 Not Found – Learner not found
            404 Not Found – SEND Category not found
    Request Headers: Authorization - JWT with Institution Admin/SEND privileges
    """
    logging.info('\n2) Add SEND category to a learner --------------------------------------')
    institution_user.send_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    # Create a new SEND Category before assign it to a learner
    str_path = '/api/v2/institution/{}/send/'.format(institution_id)
    data = {'enabled_options': ['text_to_speech', 'big_fonts', 'high_contrast'],
            'disabled_instruments': [1, 2, 3, 4, 5]}
    str_data = {'parent': '',
                'description': 'SEND Category for TESTING purposes',
                'data': data}
    new_send_category_id = tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                                            'Create a new SEND Category',
                                                            'RESPONSE: ', 201)

    # Add the new SEND category to tester learner user.
    str_path = '/api/v2/institution/{}/learner/{}/send/'.format(institution_id, learner_user_id)
    str_data = {'category': new_send_category_id}
    tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                     'Add SEND category to a learner',
                                     'RESPONSE: ', 201)
    # Assert SEND categories has increased.
    str_path = '/api/v2/institution/{}/learner/{}/send/'.format(institution_id, learner_user_id)
    str_response = 'RESPONSE Learner ID={}:'.format(learner_user_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'Read SEND categories assigned to a learner',
                                           str_response, 200)
    assert body['count'] == 1
    new_send_assignment_id = body['results'][0]['id']

    # 3) Remove a SEND Category from a learner
    """ ---------------------------------------------------------------------
    Remove a SEND Category from a learner:
        DELETE /api/v2/institution/(int: institution_id)/learner/(int: learner_id)/send/(int: send_assig_id)/
        Status Codes:
            200 OK – Ok
            400 Bad Request – Invalid information provided. The response contains the description of the errors.
            404 Not Found – Institution not found
            404 Not Found – Learner not found
            404 Not Found – SEND Category not found
    Request Headers: Authorization - JWT with Institution Admin/SEND privileges
    """
    logging.info('\n3) Remove a SEND Category from a learner --------------------------------------')
    institution_user.send_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_path = '/api/v2/institution/{}/learner/{}/send/{}/'.format(institution_id,
                                                                   learner_user_id,
                                                                   new_send_assignment_id)
    tests.utils.delete_rest_api_client(rest_api_client, str_path,
                                       'Remove a SEND Category from a learner', "RESPONSE: ", 204)

    # Delete SEND Category for testing purposes
    str_path = '/api/v2/institution/{}/send/{}/'.format(institution_id, new_send_category_id)
    tests.utils.delete_rest_api_client(rest_api_client, str_path,
                                       'Delete SEND Category', "RESPONSE: ", 204)

    str_path = '/api/v2/institution/{}/learner/{}/send/'.format(institution_id, learner_user_id)
    str_response = 'RESPONSE Learner ID={}:'.format(learner_user_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'Read SEND categories assigned to a learner',
                                           str_response, 200)
    assert body['count'] == 0



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
""" Vault Manager Module"""
import base64
import binascii
import datetime
import json
from typing import Union

import hvac
import pytz
from django.utils import timezone
from requests.exceptions import ConnectionError

from .setup import VaultSetup
from ..config import ConfigManager
from ..exception import TeslaVaultException
from ..modules import get_modules


class VaultManager:
    """ Manager class for Vault """

    #: Vault url
    vault_url: Union[str, None] = None

    def __init__(self, config=None):
        """
            Default constructor
            :param config: Configuration manager instance
            :type config: ConfigManager
        """
        if config is None:
            config = ConfigManager()

        # Store the configuration object
        self._config = config

        # Check Vault URL
        self.vault_url = self._config.config.get('VAULT_URL')
        if not self.vault_url.startswith('http://') and not self.vault_url.startswith('https://'):
            raise TeslaVaultException('Invalid VAULT_URL value. Schema is expected.')

        # Initialize the Vault client
        self._client = hvac.Client(url=self.vault_url, verify=config.config.get('VAULT_SSL_VERIFY'))

        # Set root token if provided and management is enabled
        token = self._config.config.get('VAULT_TOKEN')
        if token is not None and self._config.config.get('VAULT_MANAGEMENT'):
            self._client.token = token

        # Set mount points
        self._approle_mount_point = self._config.config.get('VAULT_MOUNT_PATH_APPROLE')
        self._transit_mount_point = self._config.config.get('VAULT_MOUNT_PATH_TRANSIT')

        # Try to authenticate
        try:
            self._auth()
            self._is_ready = True
            self._error = None
        except TeslaVaultException as exc:
            self._is_ready = False
            self._error = exc

    def _module_auth(self):
        """
            Authenticate with vault using module credentials
        """
        # Authenticate with role and secret
        if not self._client.is_authenticated():
            role_id = self._config.config.get('VAULT_ROLE_ID')
            secret_id = self._config.config.get('VAULT_SECRET_ID')
            if role_id is None or secret_id is None:
                raise TeslaVaultException('Missing Vault module credentials')
            if isinstance(role_id, list) and len(role_id) != 1:
                raise TeslaVaultException('Multiple credentials not allowed on Vault Manager')
            self._client.auth.approle.login(role_id[0], secret_id[0],
                                            mount_point=self._config.config.get('VAULT_MOUNT_PATH_APPROLE'))

        # Check authentication
        if not self._client.is_authenticated():
            raise TeslaVaultException('Cannot authenticate with Vault')

    def _auth(self):
        """
            Perform authentication with Vault
        """
        try:
            if not self._client.sys.is_initialized() or self._client.sys.is_sealed():
                if not self._config.config.get('VAULT_MANAGEMENT'):
                    # Only management modules are able to work with not initialized or sealed vault instances
                    if not self._client.sys.is_initialized():
                        raise TeslaVaultException('Vault is not initialized when management is disabled.')
                    if self._client.sys.is_sealed():
                        raise TeslaVaultException('Vault is sealed when management is disabled.')
            else:
                # Management is disabled. Use module authentication.
                self._module_auth()
        except ConnectionError as conn_err:
            raise TeslaVaultException('Vault is not responding at {}.'.format(self.vault_url)) from conn_err

    def initialize(self):
        """
            Initialize Vault server
        """
        if not self._client.sys.is_initialized():
            # Initialize vault
            result = self._client.sys.initialize(5, 2)
            root_token = result['root_token']
            keys = result['keys']
            # Store token and keys
            self._config.config.set('VAULT_TOKEN', root_token)
            self._config.config.set('VAULT_KEYS', keys)
            # Store the credentials
            self._config.save_configuration()
            # Store token on client
            self._client.token = root_token

        if self._client.sys.is_sealed():
            # Unseal vault
            self.unseal(self._config.config.get('VAULT_KEYS'))

        if not self._client.sys.is_initialized() or self._client.sys.is_sealed():
            raise TeslaVaultException('Failed to initialize and unseal vault')

        # Setup vault information to be used in TeSLA
        self._setup_vault()

    def unseal(self, keys=None):
        """
            Unseal Vault server

            :param keys: Unseal keys for vault
        """
        if not self._client.sys.is_sealed():
            return
        # Get unseal keys
        if keys is None:
            keys = self._config.config.get('VAULT_KEYS')
        if keys is None:
            raise TeslaVaultException('Unseal keys not provided')
        self._client.sys.submit_unseal_keys(keys=keys)

    def _setup_vault(self):
        """
            Prepare Vault to be used by TeSLA. Creates the roles, configurations and policies.
        """
        # Create the setup object
        setup = VaultSetup(self._client, self._config.config)

        # Setup vault
        setup.run_setup()

    def get_module_credentials(self, module):
        """
            Generate new credentials for a given module

            :param module: Module name
            :type module: str
            :return: Object with the role_id and secret_id values
            :rtype: dict
        """
        # Check the module name
        if module not in get_modules() and not module.startswith('vle_') and not module.startswith('provider_'):
            raise TeslaVaultException('Invalid module name {}'.format(module))

        # Generate the new credentials
        role_id = self._client.auth.approle.read_role_id(module, mount_point=self._approle_mount_point)
        secret_id = self._client.auth.approle.generate_secret_id(module, mount_point=self._approle_mount_point)

        return {
            'role_id': role_id['data']['role_id'],
            'secret_id': secret_id['data']['secret_id']
        }

    def is_initialized(self):
        """
            Check if Vault is initialized

            :return: True if Vault is initialized or False otherwise
            :rtype: bool
        """
        return self._client.sys.is_initialized()

    def is_sealed(self):
        """
            Check if Vault is sealed

            :return: True if Vault is sealed or False otherwise
            :rtype: bool
        """
        return self._client.sys.is_sealed()

    def is_ready(self):
        """
            Check if Vault is ready to be used

            :return: True if Vault is ready or False otherwise
            :rtype: bool
        """
        return self.is_initialized() and not self.is_sealed()

    def create_token(self, data, ttl=120, key='default'):
        """
            Generates a JWT token that can be validated using VaultManager.

            :param data: Data to include in the token
            :type data: dict
            :param ttl: Validity time in minutes
            :type ttl: int
            :param key: The JWT key to be used for token generation
            :type key: str
            :return: JWT token
            :rtype: str
        """
        # Get current key
        key_name = 'jwt_{}'.format(key)
        try:
            jwt_signing_key = self._client.secrets.transit.read_key(name=key_name,
                                                                    mount_point=self._transit_mount_point)
            jwt_signing_key = jwt_signing_key['data']
        except hvac.exceptions.Forbidden:
            raise TeslaVaultException('Module not allowed to access JWT signing key {}'.format(key_name))

        # Build the header
        header = {
            'alg': "RS256",
            'typ': "JWT",
            'ver': '{}:v{}'.format(key, jwt_signing_key['latest_version'])
        }

        # Get current time and expiration
        current_time = timezone.now()
        expiration_time = current_time + timezone.timedelta(minutes=ttl)

        # Build payload with basic fields
        payload = {
            'iss': self.vault_url,
            'iat': int(current_time.timestamp()),
            'exp': int(expiration_time.timestamp()),
            'group': key,
        }

        # Add extra arguments to payload
        if data is not None:
            for arg in data:
                payload[arg] = data[arg]

        # Create the token
        b64_header = base64.urlsafe_b64encode(json.dumps(header).encode('utf8')).decode('ascii')
        b64_payload = base64.urlsafe_b64encode(json.dumps(payload).encode('utf8')).decode('ascii')
        b64_message = base64.b64encode('{}.{}'.format(b64_header, b64_payload).encode('utf8')).decode('ascii')

        # Sign the token
        try:
            sign_data_response = self._client.secrets.transit.sign_data(
                name=key_name,
                hash_input=b64_message,
                hash_algorithm='sha2-256',
                signature_algorithm='pkcs1v15',
                key_version=jwt_signing_key['latest_version'],
                mount_point=self._transit_mount_point,
            )
        except hvac.exceptions.Forbidden:
            raise TeslaVaultException('Module not allowed to sign JWT tokens with key {}'.format(key_name))

        signature = sign_data_response['data']['signature']

        # Extract the signature
        jws = signature.split(':')[2]
        b64_jws = base64.urlsafe_b64encode(jws.encode('utf8')).decode('ascii')

        # Build the token
        token = '{}.{}.{}'.format(b64_header, b64_payload, b64_jws)

        return token

    def validate_token(self, token):
        """
            Validate a JWT token

            :param token: Token to be validated
            :type token: str
            :return: Token validation data
            :rtype: dict
        """

        # Get token parts
        token_parts = token.split('.')

        if len(token_parts) != 3:
            raise TeslaVaultException('Invalid JWT format')

        # Get the data from the token
        try:
            header = json.loads(base64.urlsafe_b64decode(token_parts[0].encode('utf8')).decode('ascii'))
        except binascii.Error:
            raise TeslaVaultException('Invalid header encoding')
        try:
            payload = json.loads(base64.urlsafe_b64decode(token_parts[1].encode('utf8')).decode('ascii'))
        except binascii.Error:
            raise TeslaVaultException('Invalid payload encoding')

        # Get message for verification
        message = '{}.{}'.format(token_parts[0], token_parts[1])

        # Get key data
        key_ver = header['ver'].split(':')
        key_name = 'jwt_{}'.format(key_ver[0])
        key_version = key_ver[1]
        try:
            signature = base64.urlsafe_b64decode(token_parts[2]).decode('ascii')
        except binascii.Error:
            raise TeslaVaultException('Invalid signature encoding')
        signature = 'vault:{}:{}'.format(key_version, signature)

        # Verify token signature
        try:
            verify_signed_data_response = self._client.secrets.transit.verify_signed_data(
                name=key_name,
                hash_input=base64.urlsafe_b64encode(message.encode('utf8')).decode('ascii'),
                signature=signature,
                hash_algorithm='sha2-256',
                signature_algorithm='pkcs1v15',
                mount_point=self._transit_mount_point,
            )
        except hvac.exceptions.Forbidden:
            raise TeslaVaultException('Module not allowed to verify JWT tokens with key {}'.format(key_name))

        valid = verify_signed_data_response['data']['valid']
        if not valid:
            raise TeslaVaultException('Invalid Token Signature')

        # Check that the group is correct
        if key_ver[0] != payload['group']:
            raise TeslaVaultException('Invalid Payload group')

        # Check if token is expired
        expired = timezone.now() >= datetime.datetime.fromtimestamp(payload['exp'], tz=pytz.utc)
        if expired:
            raise TeslaVaultException('Token expired')

        return {
            'valid': valid,
            'payload': payload
        }


    def _build_jwt_token(self, key_name, key_version, header, payload):
        """
            Build a JWT token

            :param key_name: Name of the cryptographic key to be used
            :type key_name: str
            :param key_version: Version of the cryptographic key to be used
            :type key_version: str
            :param header: Token header data
            :type header: dict
            :param payload: Token payload data
            :type payload: dict
            :return: JWT token
        """

        # Create the token
        b64_header = base64.urlsafe_b64encode(json.dumps(header).encode('utf8')).decode('ascii')
        b64_payload = base64.urlsafe_b64encode(json.dumps(payload).encode('utf8')).decode('ascii')
        b64_message = base64.b64encode('{}.{}'.format(b64_header, b64_payload).encode('utf8')).decode('ascii')

        # Sign the token
        try:
            sign_data_response = self._client.secrets.transit.sign_data(
                name=key_name,
                hash_input=b64_message,
                hash_algorithm='sha2-256',
                signature_algorithm='pkcs1v15',
                key_version=key_version,
                mount_point=self._transit_mount_point,
            )
        except hvac.exceptions.Forbidden:
            raise TeslaVaultException('Module not allowed to sign JWT tokens with key {}'.format(key_name))

        signature = sign_data_response['data']['signature']

        # Extract the signature
        jws = signature.split(':')[2]
        b64_jws = base64.urlsafe_b64encode(jws.encode('utf8')).decode('ascii')

        # Build the token
        token = '{}.{}.{}'.format(b64_header, b64_payload, b64_jws)

        return token

    def refresh_token(self, access_token, refresh_token):
        """
            Refresh given token.

            :param access_token: The JWT token to be refreshed
            :type access_token: str
            :param refresh_token: The JWT refresh token authenticating this operation
            :type refresh_token: str
            :return: JWT token
            :rtype: str
        """
        # Verify token validity
        refresh_token_payload_resp = self.validate_token(refresh_token)
        refresh_token_payload = refresh_token_payload_resp['payload']

        # Get token parts
        refresh_token_parts = refresh_token.split('.')
        access_token_parts = access_token.split('.')

        refresh_token_header = json.loads(base64.b64decode(refresh_token_parts[0]))
        access_token_header = json.loads(base64.b64decode(access_token_parts[0]))
        access_token_payload = json.loads(base64.b64decode(access_token_parts[1]))

        # Check tokens compatibility
        if access_token_payload['type'] != refresh_token_payload['type'] \
            or access_token_payload['pk'] != refresh_token_payload['pk'] \
            or access_token_header['ver'].split(':')[0] != refresh_token_header['ver'].split(':')[0]:
            raise TeslaVaultException('Invalid refresh token')

        new_payload = access_token_payload
        del new_payload['iss']
        del new_payload['iat']
        del new_payload['exp']

        return self.create_token(new_payload, refresh_token_payload['ttl'], refresh_token_header['ver'].split(':')[0])

    def create_token_pair(self, data, ttl=15, max_ttl=120, key='default'):
        """
            Generates a JWT token pair that can be validated using VaultManager.

            :param data: Data to include in the token
            :type data: dict
            :param ttl: Validity time in minutes
            :type ttl: int
            :param max_ttl: Maximum time this token can be refreshed
            :type max_ttl: int
            :param key: The JWT key to be used for token generation
            :type key: str
            :return: JWT token
            :rtype: str
        """
        # Get current key
        key_name = 'jwt_{}'.format(key)
        try:
            jwt_signing_key = self._client.secrets.transit.read_key(name=key_name,
                                                                    mount_point=self._transit_mount_point)
            jwt_signing_key = jwt_signing_key['data']
        except hvac.exceptions.Forbidden:
            raise TeslaVaultException('Module not allowed to access JWT signing key {}'.format(key_name))

        # Build the header
        header = {
            'alg': "RS256",
            'typ': "JWT",
            'ver': '{}:v{}'.format(key, jwt_signing_key['latest_version'])
        }

        # Get current time and expiration
        current_time = timezone.now()
        expiration_time = current_time + timezone.timedelta(minutes=ttl)

        # Build payload with basic fields
        payload = {
            'iss': self.vault_url,
            'iat': int(current_time.timestamp()),
            'exp': int(expiration_time.timestamp()),
            'group': key,
        }

        # Add extra arguments to payload
        if data is not None:
            for arg in data:
                payload[arg] = data[arg]

        # Create the token
        access_token = self._build_jwt_token(key_name, jwt_signing_key['latest_version'], header, payload)

        # Fix the header and scope for the refresh token
        refresh_expiration_time = current_time + timezone.timedelta(minutes=max_ttl)
        payload['exp'] = int(refresh_expiration_time.timestamp())
        payload['scope'] = '/api/v2/auth/token/refresh'
        payload['ttl'] = ttl
        refresh_token = self._build_jwt_token(key_name, jwt_signing_key['latest_version'], header, payload)

        return {
            'access_token': access_token,
            'refresh_token': refresh_token
        }

    def verify_user_password(self, email, password):
        """
            Verify user credentials
        """
        # TODO: Check password on Vault
        if not email == password:
            raise TeslaVaultException('Invalid credentials')

        return {
            'email': email
        }

    def register_vle(self, vle):
        """
            Register or update information for a VLE

            :param vle: VLE model instance
            :type vle: tesla_ce.models.VLE
            :return: VLE connection options
            :rtype: dict
        """
        # Create the setup object
        setup = VaultSetup(self._client, self._config.config)

        # Create the new entity
        try:
            module_name = setup.register_vle(vle)

            # Generate new credentials for the VLE
            credentials = self.get_module_credentials(module_name)
            return {
                'role_id': credentials['role_id'],
                'secret_id': credentials['secret_id']
            }
        except hvac.exceptions.Forbidden as ex:
            raise TeslaVaultException('Operation not allowed', ex)

    def register_provider(self, provider):
        """
            Register or update information for a Provider

            :param provider: Provider model instance
            :type provider: tesla_ce.models.Provider
            :return: Provider connection options
            :rtype: dict
        """
        # Create the setup object
        setup = VaultSetup(self._client, self._config.config)

        # Create the new entity
        try:
            module_name = setup.register_provider(provider)

            # Generate new credentials for the Provider
            credentials = self.get_module_credentials(module_name)
            return {
                'role_id': credentials['role_id'],
                'secret_id': credentials['secret_id']
            }
        except hvac.exceptions.Forbidden:
            raise TeslaVaultException('Operation not allowed')

    def get_status(self):
        """
            Get the vault status

            :return: Object with status information
            :rtype: dict
        """
        return {
            'status': 0,
            'warnings': 0,
            'errors': 0,
            'info': {}
        }



# Generated by Django 3.2.4 on 2021-09-07 14:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0018_assessmentsession_auto_closed'),
    ]

    operations = [
        migrations.AddField(
            model_name='institution',
            name='allow_learner_audit',
            field=models.BooleanField(default=False, help_text='Learners can access the audit data of their reports', null=None),
        ),
        migrations.AddField(
            model_name='institution',
            name='allow_learner_report',
            field=models.BooleanField(default=False, help_text='Learners can access their reports', null=None),
        ),
        migrations.AddField(
            model_name='institution',
            name='allow_valid_audit',
            field=models.BooleanField(default=False, help_text='Audit data is available even when results are valid', null=None),
        ),
        migrations.AddField(
            model_name='reportactivityinstrument',
            name='audit_data',
            field=models.FileField(help_text='Path to the audit data.', max_length=250, null=True, upload_to=''),
        ),
    ]



# Generated by Django 3.0.6 on 2020-05-06 11:42

import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='HistogramLearnerProvider',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('b0', models.BigIntegerField(default=0, help_text='Bin containing values from 0 to 9%')),
                ('b1', models.BigIntegerField(default=0, help_text='Bin containing values from 10% to 19%')),
                ('b2', models.BigIntegerField(default=0, help_text='Bin containing values from 20% to 29%')),
                ('b3', models.BigIntegerField(default=0, help_text='Bin containing values from 30% to 39%')),
                ('b4', models.BigIntegerField(default=0, help_text='Bin containing values from 40% to 49%')),
                ('b5', models.BigIntegerField(default=0, help_text='Bin containing values from 50% to 59%')),
                ('b6', models.BigIntegerField(default=0, help_text='Bin containing values from 60% to 69%')),
                ('b7', models.BigIntegerField(default=0, help_text='Bin containing values from 70% to 79%')),
                ('b8', models.BigIntegerField(default=0, help_text='Bin containing values from 80% to 89%')),
                ('b9', models.BigIntegerField(default=0, help_text='Bin containing values from 90% to 100%')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('learner', models.ForeignKey(help_text='Related Learner.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Learner')),
                ('provider', models.ForeignKey(help_text='Related Provider.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Provider')),
            ],
            options={
                'unique_together': {('learner', 'provider')},
            },
        ),
        migrations.CreateModel(
            name='HistogramLearnerInstrument',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('b0', models.BigIntegerField(default=0, help_text='Bin containing values from 0 to 9%')),
                ('b1', models.BigIntegerField(default=0, help_text='Bin containing values from 10% to 19%')),
                ('b2', models.BigIntegerField(default=0, help_text='Bin containing values from 20% to 29%')),
                ('b3', models.BigIntegerField(default=0, help_text='Bin containing values from 30% to 39%')),
                ('b4', models.BigIntegerField(default=0, help_text='Bin containing values from 40% to 49%')),
                ('b5', models.BigIntegerField(default=0, help_text='Bin containing values from 50% to 59%')),
                ('b6', models.BigIntegerField(default=0, help_text='Bin containing values from 60% to 69%')),
                ('b7', models.BigIntegerField(default=0, help_text='Bin containing values from 70% to 79%')),
                ('b8', models.BigIntegerField(default=0, help_text='Bin containing values from 80% to 89%')),
                ('b9', models.BigIntegerField(default=0, help_text='Bin containing values from 90% to 100%')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('instrument', models.ForeignKey(help_text='Related Instrument.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Instrument')),
                ('learner', models.ForeignKey(help_text='Related Learner.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Learner')),
            ],
            options={
                'unique_together': {('learner', 'instrument')},
            },
        ),
        migrations.CreateModel(
            name='HistogramActivityProvider',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('b0', models.BigIntegerField(default=0, help_text='Bin containing values from 0 to 9%')),
                ('b1', models.BigIntegerField(default=0, help_text='Bin containing values from 10% to 19%')),
                ('b2', models.BigIntegerField(default=0, help_text='Bin containing values from 20% to 29%')),
                ('b3', models.BigIntegerField(default=0, help_text='Bin containing values from 30% to 39%')),
                ('b4', models.BigIntegerField(default=0, help_text='Bin containing values from 40% to 49%')),
                ('b5', models.BigIntegerField(default=0, help_text='Bin containing values from 50% to 59%')),
                ('b6', models.BigIntegerField(default=0, help_text='Bin containing values from 60% to 69%')),
                ('b7', models.BigIntegerField(default=0, help_text='Bin containing values from 70% to 79%')),
                ('b8', models.BigIntegerField(default=0, help_text='Bin containing values from 80% to 89%')),
                ('b9', models.BigIntegerField(default=0, help_text='Bin containing values from 90% to 100%')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('activity', models.ForeignKey(help_text='Related Activity.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Activity')),
                ('provider', models.ForeignKey(help_text='Related Provider.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Provider')),
            ],
            options={
                'unique_together': {('activity', 'provider')},
            },
        ),
        migrations.CreateModel(
            name='HistogramActivityInstrument',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('b0', models.BigIntegerField(default=0, help_text='Bin containing values from 0 to 9%')),
                ('b1', models.BigIntegerField(default=0, help_text='Bin containing values from 10% to 19%')),
                ('b2', models.BigIntegerField(default=0, help_text='Bin containing values from 20% to 29%')),
                ('b3', models.BigIntegerField(default=0, help_text='Bin containing values from 30% to 39%')),
                ('b4', models.BigIntegerField(default=0, help_text='Bin containing values from 40% to 49%')),
                ('b5', models.BigIntegerField(default=0, help_text='Bin containing values from 50% to 59%')),
                ('b6', models.BigIntegerField(default=0, help_text='Bin containing values from 60% to 69%')),
                ('b7', models.BigIntegerField(default=0, help_text='Bin containing values from 70% to 79%')),
                ('b8', models.BigIntegerField(default=0, help_text='Bin containing values from 80% to 89%')),
                ('b9', models.BigIntegerField(default=0, help_text='Bin containing values from 90% to 100%')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('activity', models.ForeignKey(help_text='Related Activity.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Activity')),
                ('instrument', models.ForeignKey(help_text='Related Instrument.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Instrument')),
            ],
            options={
                'unique_together': {('activity', 'instrument')},
            },
        ),
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Provider api serialize module."""
import jsonschema

from rest_framework import exceptions
from rest_framework import serializers

from tesla_ce import get_default_client
from tesla_ce.apps.api.utils import decode_json
from tesla_ce.apps.api.utils import JSONField
from tesla_ce.apps.api.utils import JSONFormField
from tesla_ce.lib.exception import TeslaVaultException
from tesla_ce.models import Instrument
from tesla_ce.models import Provider
from .instrument import InstrumentSerializer


class ProviderSerializer(serializers.ModelSerializer):
    """Provider serialize model module."""
    _credentials = None

    id = serializers.ReadOnlyField()
    instrument = InstrumentSerializer(read_only=True)
    instrument_id = serializers.HiddenField(default=None, allow_null=True)
    options_schema = JSONField(allow_null=True, default=None)
    options = JSONFormField(allow_null=True, default=None, schema='options_schema')
    enabled = serializers.BooleanField(allow_null=True, default=False)
    validation_active = serializers.BooleanField(allow_null=True, default=False)

    credentials = serializers.SerializerMethodField(default=None, read_only=True)

    class Meta:
        model = Provider
        fields = ['id', 'instrument_id', 'instrument', 'name', 'queue', 'description', 'url', 'version', 'acronym',
                  'allow_validation', 'inverted_polarity', 'image', 'has_service', 'service_port', 'options_schema',
                  'options', 'credentials', 'enabled', 'validation_active']

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['instrument_id'] = self.context['view'].kwargs['parent_lookup_instrument_id']
        attrs['instrument'] = Instrument.objects.get(id=attrs['instrument_id'])

        # Check options and schema
        if 'options' in attrs or 'options_schema' in attrs:
            options_schema = attrs.get('options_schema')
            if options_schema is None and 'options_schema' not in attrs and self.instance is not None:
                options_schema = decode_json(self.instance.options_schema)
            options = attrs.get('options')
            if options is None and 'options' not in attrs and self.instance is not None:
                options = decode_json(self.instance.options)
            if options_schema is None and options is not None:
                raise serializers.ValidationError(detail="Options are not compatible with the NULL provided schema")
            if options is not None:
                try:
                    jsonschema.validate(instance=options, schema=options_schema)
                except jsonschema.exceptions.ValidationError as val_err:
                    raise serializers.ValidationError(detail=val_err) from val_err

        # Apply validators
        for validator in self.get_validators():
            validator(attrs, self)
        return super().validate(attrs)

    def create(self, validated_data):
        provider = super().create(validated_data)

        try:
            provider_info = get_default_client().register_provider(validated_data, force_update=True)
        except TeslaVaultException:
            provider.delete()
            raise exceptions.PermissionDenied('Cannot register Provider')

        self._credentials = provider_info

        return provider

    def get_credentials(self, object):
        return self._credentials



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Module for Authentication API profile """
import pytest
from tests.utils import get_profile


def check_roles(profile, roles=None):

    assert 'roles' in profile

    available_roles = ['GLOBAL_ADMIN', 'LEARNER', 'INSTRUCTOR', 'ADMIN', 'SEND', 'LEGAL', 'DATA']

    if roles is None:
        roles = []

    for role in roles:
        available_roles.remove(role)

    for role in roles:
        assert role in profile['roles']
        if 'institution' in profile and profile['institution'] is not None:
            if role == 'GLOBAL_ADMIN':
                assert role not in profile['institution']['roles']
            else:
                assert role in profile['institution']['roles']

    for role in available_roles:
        assert role not in profile['roles']
        if 'institution' in profile and profile['institution'] is not None:
            assert role not in profile['institution']['roles']


@pytest.mark.django_db
def test_api_authentication_profile(rest_api_client, api_client, institution_course_test_case,
                                    providers, user_global_admin):

    # Test profile without authentication
    noauth_resp = rest_api_client.get('/api/v2/auth/profile')
    assert noauth_resp.status_code == 401

    # Get vle credentials
    vle_token = api_client.get_module_token(
        scope={},
        module_id='vle_{}'.format(str(institution_course_test_case['vle'].id).zfill(3))
    )

    vle_profile_resp = get_profile(rest_api_client, vle_token)
    assert vle_profile_resp.status_code == 403

    # Get provider credentials
    provider_token = api_client.get_module_token(
        scope={},
        module_id='provider_{}'.format(str(providers['fr'].id).zfill(3))
    )

    vle_profile_resp = get_profile(rest_api_client, provider_token)
    assert vle_profile_resp.status_code == 403

    # Get global admin token
    global_admin_token = api_client.get_admin_token_pair(user_global_admin, {})
    assert global_admin_token is not None
    assert 'access_token' in global_admin_token
    global_admin_profile_resp = get_profile(rest_api_client, global_admin_token['access_token'])
    assert global_admin_profile_resp.status_code == 200
    check_roles(global_admin_profile_resp.data, roles=['GLOBAL_ADMIN'])

    # Get institution user token
    inst_user_token = api_client.get_user_token_pair(institution_course_test_case['user'].institutionuser, {})
    assert inst_user_token is not None
    assert 'access_token' in inst_user_token
    inst_user_profile_resp = get_profile(rest_api_client, inst_user_token['access_token'])
    assert inst_user_profile_resp.status_code == 200
    assert 'institution' in inst_user_profile_resp.data
    check_roles(inst_user_profile_resp.data, roles=[])

    # Get institution admin token
    admin_user = institution_course_test_case['user'].institutionuser
    admin_user.inst_admin = True
    admin_user.data_admin = True
    admin_user.legal_admin = True
    admin_user.send_admin = True
    admin_user.save()

    inst_admin_user_token = api_client.get_user_token_pair(admin_user, {})
    assert inst_admin_user_token is not None
    assert 'access_token' in inst_admin_user_token
    inst_admin_user_profile_resp = get_profile(rest_api_client, inst_admin_user_token['access_token'])
    assert inst_admin_user_profile_resp.status_code == 200
    assert 'institution' in inst_admin_user_profile_resp.data
    check_roles(inst_admin_user_profile_resp.data, roles=['ADMIN', 'SEND', 'LEGAL', 'DATA'])

    # Get institution learner token
    learner_user = institution_course_test_case['learner'].institutionuser.learner

    inst_learner_user_token = api_client.get_user_token_pair(learner_user, {})
    assert inst_learner_user_token is not None
    assert 'access_token' in inst_learner_user_token
    inst_learner_user_profile_resp = get_profile(rest_api_client, inst_learner_user_token['access_token'])
    assert inst_learner_user_profile_resp.status_code == 200
    assert 'institution' in inst_learner_user_profile_resp.data
    check_roles(inst_learner_user_profile_resp.data, roles=['LEARNER'])

    # Get institution instructor token
    instructor_user = institution_course_test_case['instructor'].institutionuser.instructor

    inst_instructor_user_token = api_client.get_user_token_pair(instructor_user, {})
    assert inst_instructor_user_token is not None
    assert 'access_token' in inst_learner_user_token
    inst_instructor_user_profile_resp = get_profile(rest_api_client, inst_instructor_user_token['access_token'])
    assert inst_instructor_user_profile_resp.status_code == 200
    assert 'institution' in inst_instructor_user_profile_resp.data
    check_roles(inst_instructor_user_profile_resp.data, roles=['INSTRUCTOR'])



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Deployment Manager Module"""
import requests
from simplejson.errors import JSONDecodeError
from .deployment import BaseDeployment


class DeploymentManager:
    """ Manager class for Deployment """

    def __init__(self, client):
        """
            Default constructor
            :param client: Client instance
            :type client: Client
        """
        # Store the client
        self._client = client

        # Store the configuration object
        self._config = client.config

        # Create an instance of the deployment object
        self._deploy = BaseDeployment.get_instance(self._client)

    def get_status(self):
        """
            Get the deployment status

            :return: Object with status information
            :rtype: dict
        """
        return self._deploy.get_status()

    def get_deployment_scripts(self, orchestrator=None):
        """
            Get the deployment scripts according to the selected orchestrator

            :param orchestrator: The orchestrator to be used
            :type orchestrator: str
            :return: Object with all scripts and required files
            :rtype: dict
        """
        # Generate the deployment scripts
        if orchestrator is not None:
            instance = BaseDeployment.get_instance(self._client, orchestrator)
            return instance.get_deployment_scripts()
        return self._deploy.get_deployment_scripts()

    def get_services_deployment_scripts(self, orchestrator=None):
        """
            Get the services deployment scripts according to the selected orchestrator

            :param orchestrator: The orchestrator to be used
            :type orchestrator: str
            :return: Object with all scripts and required files
            :rtype: dict
        """
        # Generate the services deployment scripts
        if orchestrator is not None:
            instance = BaseDeployment.get_instance(self._client, orchestrator)
            return instance.get_services_scripts()
        return self._deploy.get_services_scripts()

    def get_vle_deployment_scripts(self, vle, orchestrator=None):
        """
            Get the vle deployment scripts according to the selected orchestrator

            :param vle: The VLE instance
            :type vle: tesla_ce.models.VLE
            :param orchestrator: The orchestrator to be used
            :type orchestrator: str
            :return: Object with all scripts and required files
            :rtype: dict
        """
        # Generate the VLE deployment scripts
        if orchestrator is not None:
            instance = BaseDeployment.get_instance(self._client, orchestrator)
            return instance.get_vle_scripts(vle)
        return self._deploy.get_vle_scripts(vle)

    def get_provider_deployment_scripts(self, provider, orchestrator=None, credentials=None):
        """
            Get the provider deployment scripts according to the selected orchestrator

            :param provider: The Provider instance
            :type provider: tesla_ce.models.Provider
            :param orchestrator: The orchestrator to be used
            :type orchestrator: str
            :param credentials: List of credentials required by this provider
            :type credentials: list
            :return: Object with all scripts and required files
            :rtype: dict
        """
        # Generate the VLE deployment scripts
        if orchestrator is not None:
            instance = BaseDeployment.get_instance(self._client, orchestrator)
            return instance.get_provider_scripts(provider, credentials)
        return self._deploy.get_provider_scripts(provider, credentials)

    @staticmethod
    def get_registered_providers(repository="tesla-ce/core", version="main"):
        """
            Get the list of registered providers

            :param repository: The repository in GitHub
            :type repository: str
            :param version: The branch on the repository
            :type version: str
            :return: List of registered providers
            :rtype: list
        """
        # Read registry of providers on the repository
        response = requests.get('https://raw.githubusercontent.com/{}/{}/providers/registry.json'.format(repository,
                                                                                                         version))
        if response.status_code != 200:
            return []

        resp_providers = response.json()

        providers = []
        for provider in resp_providers:
            provider_info_resp = requests.get(provider['url'])
            if provider_info_resp.status_code == 200:
                try:
                    providers.append(provider_info_resp.json())
                except JSONDecodeError:
                    # When fails to parse, do not include in the list
                    pass

        return providers



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for providers administration """
import pytest

import mock

from tests.conftest import get_random_string


@pytest.mark.django_db
def test_api_admin_providers(rest_api_client, user_global_admin, institution_course_test_case):
    # Ensure that only GLOBAL_ADMINS have access
    users = [
        None,
        institution_course_test_case['user'],
        institution_course_test_case['instructor'],
        institution_course_test_case['learner']
    ]
    for instrument in range(4):
        for user in users:
            rest_api_client.force_authenticate(user=user)
            no_auth_resp = rest_api_client.get('/api/v2/admin/instrument/{}/provider/'.format(instrument))
            assert no_auth_resp.status_code == 403

    # User GLOBAL_ADMIN
    for instrument in range(4):
        rest_api_client.force_authenticate(user=user_global_admin)
        valid_auth_resp = rest_api_client.get('/api/v2/admin/instrument/{}/provider/'.format(instrument))
        assert valid_auth_resp.status_code == 200

    # Register a new provider
    rest_api_client.force_authenticate(user=user_global_admin)
    prov_acronym = 'fr_tfr_{}'.format(get_random_string(5))
    prov_data = {
        'acronym':  prov_acronym,
        'version': '0.0.5',
        'url': 'https://providers.tesla-project.eu/{}'.format(prov_acronym),
        'image': 'teslace/provider-{}'.format(prov_acronym),
        'name': 'TeSLA CE TEST PROVIDER',
        'description': 'A test provider for TeSLA CE automated tests',
        'has_service': False,
        'service_port': None,
        'options_schema': {
            'type': 'object',
            'additionalProperties': False,
            'properties': {
                'option1': {
                    'type': 'number', 'default': 1
                },
                'option2': {
                        'type': 'string', 'default': 'default value'
                },
            }
        },
        'options': {
            'option1': 3
        },
        'queue': prov_acronym,
        'allow_validation': True,
        'alert_below': 0.3,
        'warning_below': 0.6,
        'inverted_polarity': False,
        'enabled': False,
        'validation_active': False
    }

    new_provider_resp = rest_api_client.post('/api/v2/admin/instrument/{}/provider/'.format(1),
                                             data=prov_data, format='json')
    assert new_provider_resp.status_code == 201
    assert 'credentials' in new_provider_resp.data
    assert new_provider_resp.data['credentials'] is not None
    assert 'role_id' in new_provider_resp.data['credentials']
    assert 'secret_id' in new_provider_resp.data['credentials']

    # Check that this provider exists
    search_provider_resp = rest_api_client.get(
        '/api/v2/admin/instrument/{}/provider/?acronym={}'.format(1, prov_acronym)
    )
    assert search_provider_resp.status_code == 200
    assert search_provider_resp.data['count'] == 1
    assert search_provider_resp.data['results'][0]['id'] == new_provider_resp.data['id']

    # Update options with a valid schema
    update_provider_resp = rest_api_client.patch(
        '/api/v2/admin/instrument/{}/provider/{}/'.format(1, new_provider_resp.data['id']),
        format='json',
        data={'options': {'option1': 35}}
    )
    assert update_provider_resp.status_code == 200
    assert update_provider_resp.data['options']['option1'] == 35

    inv_update_provider_resp = rest_api_client.patch(
        '/api/v2/admin/instrument/{}/provider/{}/'.format(1, new_provider_resp.data['id']),
        format='json',
        data={'options': {'option1': 'a'}}
    )
    assert inv_update_provider_resp.status_code == 400
    assert 'non_field_errors' in inv_update_provider_resp.data
    assert len(inv_update_provider_resp.data['non_field_errors']) == 1
    assert 'option1' in inv_update_provider_resp.data['non_field_errors'][0]

    inv2_update_provider_resp = rest_api_client.patch(
        '/api/v2/admin/instrument/{}/provider/{}/'.format(1, new_provider_resp.data['id']),
        format='json',
        data={'options': {'option3': 'a'}}
    )
    assert inv2_update_provider_resp.status_code == 400
    assert 'non_field_errors' in inv2_update_provider_resp.data
    assert len(inv2_update_provider_resp.data['non_field_errors']) == 1
    assert 'option3' in inv2_update_provider_resp.data['non_field_errors'][0]

    inv3_update_provider_resp = rest_api_client.patch(
        '/api/v2/admin/instrument/{}/provider/{}/'.format(1, new_provider_resp.data['id']),
        format='json',
        data={'options_schema': {
            'type': 'object',
            'additionalProperties': False,
            'properties': {
                'option51': {
                    'type': 'number', 'default': 1
                },
                'option52': {
                    'type': 'string', 'default': 'default value'
                },
            }
        }
    })
    assert inv3_update_provider_resp.status_code == 400
    assert 'non_field_errors' in inv3_update_provider_resp.data
    assert len(inv3_update_provider_resp.data['non_field_errors']) == 1
    assert 'option1' in inv3_update_provider_resp.data['non_field_errors'][0]

    inv4_update_provider_resp = rest_api_client.patch(
        '/api/v2/admin/instrument/{}/provider/{}/'.format(1, new_provider_resp.data['id']),
        format='json',
        data={'options_schema': None })
    assert inv4_update_provider_resp.status_code == 400
    assert 'non_field_errors' in inv4_update_provider_resp.data
    assert len(inv4_update_provider_resp.data['non_field_errors']) == 1
    assert 'NULL' in inv4_update_provider_resp.data['non_field_errors'][0]


def make_fail(self, provider_info, force_update=False):
    from tesla_ce.lib.exception import TeslaVaultException
    raise TeslaVaultException('test exception')


@pytest.mark.django_db
def test_api_admin_providers_fail(rest_api_client, user_global_admin):

    # Ensure provider is not registered in case of exception
    rest_api_client.force_authenticate(user=user_global_admin)
    prov_acronym = 'fr_tfr_{}'.format(get_random_string(5))
    prov_data = {
        'acronym':  prov_acronym,
        'version': '0.0.5',
        'url': 'https://providers.tesla-project.eu/{}'.format(prov_acronym),
        'image': 'teslace/provider-{}'.format(prov_acronym),
        'name': 'TeSLA CE TEST PROVIDER',
        'description': 'A test provider for TeSLA CE automated tests',
        'has_service': False,
        'service_port': None,
        'options_schema': {
            'type': 'object',
            'additionalProperties': False,
            'properties': {
                'option1': {
                    'type': 'number', 'default': 1
                },
                'option2': {
                        'type': 'string', 'default': 'default value'
                },
            }
        },
        'options': {
            'option1': 3
        },
        'queue': prov_acronym,
        'allow_validation': True,
        'alert_below': 0.3,
        'warning_below': 0.6,
        'inverted_polarity': False,
        'enabled': False,
        'validation_active': False
    }

    with mock.patch('tesla_ce.client.Client.register_provider', make_fail):
        new_provider_resp = rest_api_client.post('/api/v2/admin/instrument/{}/provider/'.format(1),
                                                 data=prov_data, format='json')
        assert new_provider_resp.status_code == 403
        assert 'detail' in new_provider_resp.data
        assert new_provider_resp.data['detail'] == 'Cannot register Provider'

        # Check that this provider does not exist
        search_provider_resp = rest_api_client.get(
            '/api/v2/admin/instrument/{}/provider/?acronym={}'.format(1, prov_acronym)
        )
        assert search_provider_resp.status_code == 200
        assert search_provider_resp.data['count'] == 0



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""RequestResult api serialize module."""
import simplejson
from django.core.files.base import ContentFile
from django.db.models import Q

from rest_framework import serializers

from tesla_ce import tasks
from tesla_ce.apps.api.utils import JSONField
from tesla_ce.models import Request
from tesla_ce.models import RequestProviderResult


class ProviderVerificationRequestSerializer(serializers.ModelSerializer):
    """Provider Request serialize model module."""

    id = serializers.IntegerField(read_only=True)
    data = serializers.FileField(read_only=True)

    class Meta:
        model = Request
        fields = ["id", "data"]


class ProviderVerificationRequestResultSerializer(serializers.ModelSerializer):

    request = ProviderVerificationRequestSerializer(read_only=True)
    learner_id = serializers.UUIDField(source='request.learner.learner_id', read_only=True)
    audit = serializers.FileField(read_only=True)
    audit_data = JSONField(required=False, write_only=True, allow_null=True, default=None)

    """Provider RequestResult serialize model module."""
    class Meta:
        model = RequestProviderResult
        fields = ["id", "request", "learner_id", "result", "status", "error_message", "code", "audit", "audit_data"]

    def update(self, instance, validated_data):
        new_instance = super().update(instance, validated_data)
        new_instance.audit.save('{}__audit.json'.format(new_instance.request.data.name),
                                ContentFile(simplejson.dumps(new_instance.audit_data).encode('utf-8')))

        # If all the providers reported their results, launch summarise task
        if RequestProviderResult.objects.filter(Q(status=0) | Q(status=7),
                                                request_id=new_instance.request_id,
                                                provider__instrument=new_instance.provider.instrument
                                                ).count() == 0:
            tasks.requests.verification.create_verification_summary.apply_async((
                    instance.request.id,
                    new_instance.provider.instrument_id,
            ))
        return new_instance



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" VLE Course Learner views module """
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import VLECourseInstructorSerializer
from tesla_ce.models import Course
from tesla_ce.models import Instructor


# pylint: disable=too-many-ancestors
class VLECourseInstructorViewSet(viewsets.ModelViewSet, NestedViewSetMixin):
    """
    API endpoint that allows activity to be viewed or edited.
    """
    model = Instructor
    serializer_class = VLECourseInstructorSerializer
    permission_classes = [
        permissions.VLEPermission
    ]
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    filterset_fields = ['uid', 'email', 'first_name', 'last_name']
    search_fields = ['uid', 'email', 'first_name', 'last_name']

    def get_queryset(self):
        if 'parent_lookup_vle_id' in self.kwargs and 'parent_lookup_course_id' in self.kwargs:
            course = Course.objects.get(id=self.kwargs['parent_lookup_course_id'])
            queryset = course.instructors
        else:
            queryset = Instructor.objects
        return queryset.all().order_by('id')

    def perform_destroy(self, instance):
        """
            Remove an instructor from current course
            :param instance: Instructor to be removed
        """
        course = Course.objects.get(id=self.kwargs['parent_lookup_course_id'])
        course.instructors.remove(instance)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA API utility module """
import json

from rest_framework import serializers
from rest_framework.fields import empty


def decode_json(value):
    """
        Decode a JSON from String
        :param value: String JSON object
        :type value: str
        :return: JSON object
        :rtype: dict
    """
    if value is not None and isinstance(value, str):
        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError:
            json_value = value.replace('\'', '"')
            json_value = json_value.replace('False', 'false').replace('True', 'true')
            json_value = json_value.replace('None', 'null')
            return json.loads(json_value)
    return value


class JSONField(serializers.JSONField):
    """ Deal with JSON fields """

    def to_representation(self, value):
        """
            Creates a representation for a JSON field
            :param value: Field value
            :return: Representation or the value
        """
        if isinstance(value, str) and len(value) == 0:
            value = None
        if value is not None and isinstance(value, str):
            return decode_json(value)
        return json.loads(json.dumps(value))

    def to_internal_value(self, data):
        return super().to_internal_value(data)


class BodySerializer(serializers.Serializer):
    """ Utility base serializer for custom request body """

    def update(self, instance, validated_data):
        """
            Method called when the represented object is modified.
            :param instance: Current instance
            :param validated_data: Validated data
            :return: Updated instance
        """
        return instance

    def create(self, validated_data):
        """
            Method called when a new object is created.
            :param validated_data: Validated data
            :return: Created instance
        """
        return None


class JSONFormField(JSONField):
    schema = None

    def __init__(self, *args, **kwargs):
        self.schema = self.schema = kwargs.pop('schema', None)

        assert self.schema is not None, "Schema parameter is required"

        super().__init__(*args, **kwargs)

    def run_validation(self, data=empty):
        valid_data = super().run_validation(data)
        return valid_data



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test package for provider API """
import pytest
from tests.utils import get_module_auth_user


@pytest.mark.django_db
def test_api_provider(rest_api_client, providers):
    # Ensure the list of providers is not accessible
    provider_list_no_auth_response = rest_api_client.get('/api/v2/provider/')
    assert provider_list_no_auth_response.status_code == 403

    # Authenticate with Provider object
    rest_api_client.force_authenticate(user=get_module_auth_user(providers['fr']))

    # Check that is not possible to list the Providers
    provider_list_auth_response = rest_api_client.get('/api/v2/provider/')
    assert provider_list_auth_response.status_code == 403

    # Get Provider information
    provider_id = providers['fr'].id
    provider_info_auth_response = rest_api_client.get('/api/v2/provider/{}/'.format(provider_id))
    assert provider_info_auth_response.status_code == 200
    assert provider_info_auth_response.data['id'] == provider_id



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Activity Report views module """
from django_filters.rest_framework import DjangoFilterBackend

from rest_framework import viewsets
from rest_framework import exceptions
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework_extensions.mixins import DetailSerializerMixin
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api.v2.serializers import InstitutionCourseActivityReportExtendedSerializer
from tesla_ce.apps.api.v2.serializers import InstitutionCourseActivityReportSerializer
from tesla_ce.apps.api import permissions

from tesla_ce.models import Course
from tesla_ce.models import ReportActivity
from tesla_ce.models.user import get_institution_user
from tesla_ce.models.user import is_global_admin


# pylint: disable=too-many-ancestors
class InstitutionCourseActivityReportViewSet(NestedViewSetMixin, DetailSerializerMixin, viewsets.ReadOnlyModelViewSet):
    """
    API endpoint that allows view activity reports.
    """
    model = ReportActivity
    queryset = ReportActivity.objects
    serializer_class = InstitutionCourseActivityReportSerializer
    serializer_detail_class = InstitutionCourseActivityReportExtendedSerializer
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    search_fields = ['learner__first_name', 'learner__last_name', 'learner__email',
                     'identity_level', 'content_level', 'integrity_level']
    filterset_fields = ['learner__first_name', 'learner__last_name', 'learner__email',
                        'identity_level', 'content_level', 'integrity_level']
    permission_classes = [
        permissions.GlobalAdminReadOnlyPermission |
        permissions.InstitutionAdminReadOnlyPermission |
        permissions.InstitutionDataAdminReadOnlyPermission |
        permissions.InstitutionCourseInstructorReadOnlyPermission |
        permissions.InstitutionCourseLearnerReadOnlyPermission
    ]

    def get_queryset(self):
        queryset = self.filter_queryset_by_parents_lookups(ReportActivity.objects)
        if not is_global_admin(self.request.user):
            inst_user = get_institution_user(self.request.user)
            if not inst_user.inst_admin and not inst_user.legal_admin and not inst_user.data_admin:
                course = Course.objects.get(vle__institution=self.kwargs['parent_lookup_activity__vle__institution_id'],
                                            id=self.kwargs['parent_lookup_activity__course_id'])
                try:
                    is_instructor = self.request.user.instructor in course.instructors.all()
                except Exception:
                    is_instructor = False
                if not is_instructor:
                    if not inst_user.institution.allow_learner_report:
                        raise exceptions.PermissionDenied('Not allowed')
                    queryset = queryset.filter(learner_id=inst_user.id)
        return queryset.all()



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" SENDLearner model module."""

from django.db import models

from .base_model import BaseModel
from .learner import Learner
from .learner import get_learner_send
from .send_category import SENDCategory


class SENDLearner(BaseModel):
    """ SENDLearner model. """
    learner = models.ForeignKey(Learner, null=False,
                                on_delete=models.CASCADE)
    category = models.ForeignKey(SENDCategory, null=False,
                                 on_delete=models.CASCADE)

    expires_at = models.DateTimeField(null=True)

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('learner', 'category'),)

    def __repr__(self):
        return "<SENDLearner(tesla_id='%r', category_id='%r')>" % (
            self.learner, self.category)

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        super().save(force_insert, force_update, using, update_fields)
        # Invalidate cached value
        get_learner_send.invalidate(self.learner_id)



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" API serializers package """
from .admin import InstitutionAdminSerializer
from .admin import InstrumentSerializer
from .admin import ProviderSerializer
from .admin import UIOptionSerializer
from .admin import UserSerializer
from .institution import InstitutionCourseActivityInstrumentSerializer
from .institution import InstitutionCourseActivityReportDetailSerializer
from .institution import InstitutionCourseActivityReportExtendedSerializer
from .institution import InstitutionCourseActivityReportLearnerSerializer
from .institution import InstitutionCourseActivityReportSerializer
from .institution import InstitutionCourseActivityReportAuditSerializer
from .institution import InstitutionCourseActivitySerializer
from .institution import InstitutionCourseActivityExtendedSerializer
from .institution import InstitutionCourseGroupCourseSerializer
from .institution import InstitutionCourseGroupSerializer
from .institution import InstitutionCourseSerializer
from .institution import InstitutionInformedConsentDocumentSerializer
from .institution import InstitutionInformedConsentSerializer
from .institution import InstitutionInstructorSerializer
from .institution import InstitutionLearnerDetailSerializer
from .institution import InstitutionLearnerICBodySerializer
from .institution import InstitutionLearnerSerializer
from .institution import InstitutionSENDCategorySerializer
from .institution import InstitutionSENDLearnerSerializer
from .institution import InstitutionSerializer
from .institution import InstitutionCourseActivityReportRequestSerializer
from .institution import InstitutionCourseLearnerSerializer
from .institution import InstitutionCourseInstructorSerializer
from .institution import InstitutionUIOptionSerializer
from .institution import InstitutionUserSerializer
from .institution import InstitutionVLESerializer
from .provider import ProviderEnrolmentSampleSerializer
from .provider import ProviderEnrolmentSampleValidationSerializer
from .provider import ProviderEnrolmentSerializer
from .provider import ProviderNotificationSerializer
from .provider import ProviderVerificationRequestResultSerializer
from .vle import VLECourseActivityInstrumentSerializer
from .vle import VLECourseActivityLearnerRequestSerializer
from .vle import VLECourseActivityLearnerSerializer
from .vle import VLECourseActivityReportExtendedSerializer
from .vle import VLECourseActivityReportSerializer
from .vle import VLECourseActivitySerializer
from .vle import VLECourseActivityAttachmentSerializer
from .vle import VLECourseInstructorSerializer
from .vle import VLECourseLearnerSerializer
from .vle import VLECourseSerializer
from .vle import VLEInstrumentSerializer
from .vle import VLELauncherBodySerializer
from .vle import VLELauncherDataSerializer
from .vle import VLENewAssessmentSessionBodySerializer
from .vle import VLENewAssessmentSessionSerializer
from .vle import VLESerializer

__all__ = [
    # Administration nested serializers
    "InstitutionAdminSerializer",
    "InstrumentSerializer",
    "ProviderSerializer",
    "UIOptionSerializer",
    "UserSerializer",
    # VLE nested serializers
    "VLESerializer",
    "VLECourseSerializer",
    "VLECourseActivitySerializer",
    "VLECourseActivityInstrumentSerializer",
    "VLECourseLearnerSerializer",
    "VLECourseActivityLearnerSerializer",
    "VLECourseActivityLearnerRequestSerializer",
    "VLECourseActivityAttachmentSerializer",
    "VLECourseInstructorSerializer",
    "VLENewAssessmentSessionSerializer",
    "VLENewAssessmentSessionBodySerializer",
    "VLECourseActivityReportSerializer",
    "VLECourseActivityReportExtendedSerializer",
    "VLEInstrumentSerializer",
    "VLELauncherBodySerializer",
    "VLELauncherDataSerializer",
    # Institution nested serializers
    "InstitutionSerializer",
    "InstitutionCourseActivitySerializer",
    "InstitutionCourseActivityExtendedSerializer",
    "InstitutionCourseActivityReportSerializer",
    "InstitutionCourseActivityReportAuditSerializer",
    "InstitutionCourseActivityReportDetailSerializer",
    "InstitutionCourseActivityReportExtendedSerializer",
    "InstitutionCourseActivityReportLearnerSerializer",
    "InstitutionCourseActivityInstrumentSerializer",
    "InstitutionCourseActivityInstrumentSerializer",
    "InstitutionCourseSerializer",
    "InstitutionCourseGroupSerializer",
    "InstitutionCourseGroupCourseSerializer",
    "InstitutionCourseLearnerSerializer",
    "InstitutionCourseInstructorSerializer",
    "InstitutionLearnerSerializer",
    "InstitutionCourseActivityReportRequestSerializer",
    "InstitutionInstructorSerializer",
    "InstitutionLearnerDetailSerializer",
    "InstitutionLearnerICBodySerializer",
    "InstitutionInformedConsentSerializer",
    "InstitutionInformedConsentDocumentSerializer",
    "InstitutionSENDCategorySerializer",
    "InstitutionSENDLearnerSerializer",
    "InstitutionVLESerializer",
    "InstitutionUIOptionSerializer",
    "InstitutionUserSerializer",
    # Provider nested serializers
    "ProviderEnrolmentSerializer",
    "ProviderEnrolmentSampleSerializer",
    "ProviderEnrolmentSampleValidationSerializer",
    "ProviderVerificationRequestResultSerializer",
    "ProviderNotificationSerializer",
]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Django URLs definition """
from django.conf import settings
from django.http import HttpResponse
from django.shortcuts import redirect
from django.templatetags.static import static
from django.urls import include
from django.urls import path
from django.urls import re_path

from tesla_ce import get_default_client


def favicon_view(request):
    """
        Allow the exportation of favicon when requested
        :param request: HTTP request
        :return: Redirects to the static URL with favicon
    """
    return redirect(static('img/favicon.ico'))


def version_view(request, module=None):
    """
        Return the version of the code
        :param request: HTTP request
        :param module: The module
        :return: Returns the version
    """
    version = get_default_client().version

    return HttpResponse('{"version": "' + version + '"}', content_type='application/json')


urlpatterns = [
    re_path(r'^favicon\.ico$', favicon_view),
    path('', include('django_prometheus.urls')),
    re_path(r'^(?P<module>\D+)/version/$', version_view),
]

if 'api' in settings.TESLA_MODULES:
    urlpatterns += [
        path('api/', include('tesla_ce.apps.api.urls')),
        path('api/webhooks/', include('tesla_ce.apps.webhooks.urls')),
    ]

if 'lapi' in settings.TESLA_MODULES:
    urlpatterns += [
        path('lapi/', include('tesla_ce.apps.lapi.urls'))
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Provider Enrolment views module """
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import NotAcceptable
from rest_framework.exceptions import NotFound
from rest_framework.exceptions import PermissionDenied
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework.views import Response
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import ProviderEnrolmentSampleSerializer
from tesla_ce.apps.api.v2.serializers import ProviderEnrolmentSerializer
from tesla_ce.models import Enrolment
from tesla_ce.models import EnrolmentSample
from tesla_ce.models import EnrolmentSampleValidation
from tesla_ce.models import Provider


# pylint: disable=too-many-ancestors
class ProviderEnrolmentViewSet(NestedViewSetMixin, viewsets.ModelViewSet):
    """
    API endpoint that allows Provider access enrolment samples.
    """
    queryset = Enrolment.objects
    lookup_field = 'learner__learner_id'
    serializer_class = ProviderEnrolmentSerializer
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    permission_classes = [
        permissions.ProviderPermission
    ]
    '''
    filterset_fields = ['activity_type', 'external_token', 'description', 'conf', 'vle']
    search_fields = ['activity_type', 'external_token', 'description', 'conf', 'vle']
    '''

    @action(detail=True, methods=['POST', ], serializer_class=ProviderEnrolmentSampleSerializer)
    def unlock(self, request, *args, **kwargs):
        """
            Unlock a locked model
        """
        try:
            model = self.get_queryset().get(learner__learner_id=kwargs['learner__learner_id'])
        except Enrolment.DoesNotExist:
            raise NotFound('No model to unlock')

        if not model.is_locked:
            raise NotAcceptable('Model is not locked')
        token = self.request.data.get('token')
        if str(model.locked_by) != str(token):
            raise PermissionDenied('Invalid unlock token')
        model.locked_by = None
        model.locked_at = None
        model.save()

        return Response('Model unlocked')

    @action(detail=True, methods=['GET', ], serializer_class=ProviderEnrolmentSampleSerializer)
    def available_samples(self, request, *args, **kwargs):
        """
            Get available samples for this enrolment model
        """
        try:
            used_samples = list(self.get_queryset()
                                .get(learner__learner_id=kwargs['learner__learner_id'])
                                .model_samples.values_list('id', flat=True).all())
        except Enrolment.DoesNotExist:
            used_samples = []
        provider = Provider.objects.get(id=kwargs['parent_lookup_provider_id'])
        sample_ids = list(EnrolmentSampleValidation.objects.filter(
            sample__learner__learner_id=kwargs['learner__learner_id'],
            provider__instrument=provider.instrument,
            status=1).exclude(sample_id__in=used_samples).values_list('sample', flat=True).all())
        samples = EnrolmentSample.objects.filter(id__in=sample_ids, status=1)

        page = self.paginate_queryset(samples)
        if page is not None:
            data = self.serializer_class(page, many=True).data
            return self.get_paginated_response(data)

        data = self.serializer_class(samples, many=True).data
        return Response(self.get_paginated_response(data))

    @action(detail=True, methods=['GET', ], serializer_class=ProviderEnrolmentSampleSerializer)
    def used_samples(self, request, *args, **kwargs):
        """
            Get used samples for this enrolment model
        """
        try:
            used_samples = list(self.get_queryset()
                                .get(learner__learner_id=kwargs['learner__learner_id'])
                                .model_samples.values_list('id', flat=True).all())
        except Enrolment.DoesNotExist:
            used_samples = []
        samples = EnrolmentSample.objects.filter(id__in=used_samples)
        page = self.paginate_queryset(samples)
        if page is not None:
            data = self.serializer_class(page, many=True).data
            return self.get_paginated_response(data)

        data = self.serializer_class(samples, many=True).data
        return Response(self.get_paginated_response(data))



# Generated by Django 3.0.6 on 2020-05-20 13:36

import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0003_providernotification'),
    ]

    operations = [
        migrations.CreateModel(
            name='Launcher',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('target', models.IntegerField(choices=[(0, 'Dashboard'), (1, 'LAPI')], default=0, help_text='Target for this launcher')),
                ('target_url', models.CharField(default=None, help_text='Url to redirect', max_length=250, null=True)),
                ('token', models.UUIDField(help_text='Token to allow authentication for this user')),
                ('token_pair', models.TextField(help_text='Pair of tokens to authenticate with target')),
                ('expires_at', models.DateTimeField(help_text='When this access token expires')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('session', models.ForeignKey(default=None, help_text='Related assessment session', null=True, on_delete=django.db.models.deletion.SET_NULL, to='tesla_ce.AssessmentSession')),
                ('user', models.ForeignKey(help_text='Related user', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.InstitutionUser')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Provider enrolment sample validation serializers module """
import json

from django.core.files.base import ContentFile
from rest_framework import serializers

from tesla_ce.apps.api.utils import JSONField
from tesla_ce.models import EnrolmentSampleValidation
from tesla_ce.tasks.requests.enrolment import create_validation_summary
from .enrolment_sample import ProviderEnrolmentSampleSerializer


class ProviderEnrolmentSampleValidationSerializer(serializers.ModelSerializer):
    """Enrolment sample validation serialize class."""

    sample_id = serializers.HiddenField(default=None)
    provider_id = serializers.HiddenField(default=None)
    provider = serializers.IntegerField(read_only=True, source="provider_id")
    sample = ProviderEnrolmentSampleSerializer(read_only=True)
    validation_info = JSONField(write_only=True, allow_null=True, default=None)
    info = serializers.FileField(read_only=True, allow_null=True, default=None)
    message_code_id = serializers.CharField(allow_null=True, default=None)
    contribution = serializers.FloatField(allow_null=True, default=None)

    class Meta:
        model = EnrolmentSampleValidation
        fields = ["id", "status", "info", "validation_info", "sample_id", "provider_id", "provider",
                  "sample", "error_message", "contribution", "message_code_id"]
        validators = [serializers.UniqueTogetherValidator(
            queryset=EnrolmentSampleValidation.objects.all(),
            fields=['sample_id', 'provider_id']
        ), ]

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['provider_id'] = self.context['view'].kwargs['parent_lookup_provider_id']
        attrs['sample_id'] = self.context['view'].kwargs['parent_lookup_sample_id']

        # Apply validators
        for validator in self.get_validators():
            validator(attrs, self)
        return super().validate(attrs)

    def update(self, instance, validated_data):
        """
            Update an enrolment sample validation
            :param instance: Current instance
            :type instance: EnrolmentSampleValidation
            :param validated_data: Validated data
            :type validated_data: dict
            :return: New instance
        """
        if validated_data['validation_info'] is None and instance.info is not None and len(instance.info.name) > 0:
            instance.info.delete()
        if validated_data['validation_info'] is not None:
            instance.info.save('{}.validation.{}'.format(instance.sample.data.name, instance.id),
                               ContentFile(json.dumps(validated_data['validation_info']).encode('utf-8')))
        new_instance = super().update(instance, validated_data)

        if 0 < new_instance.status < 4:
            create_validation_summary.apply_async((new_instance.sample.learner.learner_id, new_instance.sample_id,))

        return new_instance



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
""" Configuration Module"""
import configparser

from ..exception import TeslaConfigException


class Config:
    """ Configuration class

        Allows to manage the configuration values and apply hierarchical configurations
    """
    #: Configuration sections
    _sections = (
        # (section_name, section_description, fields)
        #   (field_name, field_description, field_type, default, options, exportable)
        ('tesla', 'Generic configuration', (
            ('config_file', 'Configuration file to be used', 'str', None, None, False),
            ('mode', 'Working mode of TeSLA', 'enum', 'production', ['config', 'production', ], False),
            ('domain', 'Base domain', 'str', None, None, True),
            ('admin_mail', 'Mail of the administrator', 'str', None, None, True),
            ('institution_name', 'Name of the institution', 'str', 'Default Institution', None, True),
            ('institution_acronym', 'Acronym of the institution', 'str', 'default', None, True),
        )),
        ('db', 'Database configuration', (
            ('host', 'Database host', 'str', 'localhost', None, True),
            ('engine', 'Database engine', 'enum', 'mysql', ['mysql', 'postgresql',], True),
            ('port', 'Database port', 'int', 3306, None, True),
            ('name', 'Database name', 'str', 'tesla', None, True),
            ('user', 'Database user', 'str', None, None, True),
            ('password', 'Database password', 'str', None, None, True),
            ('root_user', 'Database root username', 'str', 'root', None, True),
            ('root_password', 'Database root password', 'str', None, None, True),
        )),
        ('vault', 'HashiCorp Vault configuration', (
            ('url', 'Vault complete URL', 'str', 'http://localhost:8200', None, True),
            ('ssl_verify', 'Verify SSL', 'bool', True, None, True),
            ('token', 'Token to authenticate', 'str', None, None, True),
            ('role_id', 'Role ID value to authenticate', 'list', None, None, False),
            ('secret_id', 'Secret ID value to authenticate', 'list', None, None, False),
            ('keys', 'Unseal keys for Vault', 'list', None, None, True),
            ('management', 'Allow Vault management', 'bool', False, None, False),
            ('db_host', 'MySQL host for vault database', 'str', None, None, True),
            ('db_port', 'MySQL port for vault database', 'int', 3306, None, True),
            ('db_name', 'MySQL database name for vault database', 'str', None, None, True),
            ('db_user', 'MySQL user for vault database', 'str', None, None, True),
            ('db_password', 'MySQL password for vault database', 'str', None, None, True),
            ('mount_path_kv', 'Mount path for KV secrets engine', 'str', 'tesla-ce/kv', None, True),
            ('mount_path_transit', 'Mount path for transit encryption engine', 'str', 'tesla-ce/transit', None, True),
            ('mount_path_approle', 'Mount path for approle auth engine', 'str', 'tesla-ce/approle', None, True),
            ('policies_prefix', 'Prefix prepended to ACL policy names', 'str', 'tesla-ce/policy/', None, True),
        )),
        ('redis', 'Redis configuration', (
            ('host', 'Redis host', 'str', 'localhost', None, True),
            ('port', 'Redis port', 'int', 6379, None, True),
            ('password', 'Redis password', 'str', None, None, True),
            ('database', 'Redis database', 'int', 1, None, True),
        )),
        ('storage', 'Storage configuration (S3 or Minio)', (
            ('url', 'Storage url', 'str', 'http://localhost:9000', None, True),
            ('bucket_name', 'Storage default private bucket', 'str', 'tesla', None, True),
            ('public_bucket_name', 'Storage default public bucket', 'str', 'tesla-public', None, True),
            ('region', 'Storage region', 'str', 'eu-west-1', None, True),
            ('access_key', 'Storage access key id', 'str', None, None, True),
            ('secret_key', 'Storage secret access key', 'str', None, None, True),
            ('ssl_verify', 'Verify SSL', 'bool', True, None, True),
        )),
        ('celery', 'Celery configuration (RabbitMQ or SQS)', (
            ('broker_protocol', 'Protocol to communicate with brocker', 'enum', 'amqp', ['amqp', 'sqs'], True),
            ('broker_user', 'User to authenticate with RabbitMQ or KEY for Amazon SQS', 'str', None, True),
            ('broker_password', 'Password to authenticate with RabbitMQ or SECRET for Amazon SQS', 'str', None, True),
            ('broker_host', 'Host for RabbitMQ', 'str', 'localhost', True),
            ('broker_region', 'Region for Amazon SQS', 'str', 'eu-west-1', True),
            ('broker_port', 'Port for RabbitMQ', 'int', 5672, True),
            ('broker_vhost', 'RabbitMQ Virtual Host', 'str', '/', True),
            ('queue_default', 'Name of the default queue used for periodic tasks', 'str', 'tesla', True),
            ('queue_enrolment', 'Name of the default queue for enrolment', 'str', 'tesla_enrolment', True),
            ('queue_enrolment_storage', 'Name of the storage queue for enrolment', 'str',
             'tesla_enrolment_storage', True),
            ('queue_enrolment_validation', 'Name of the validation queue for enrolment', 'str',
             'tesla_enrolment_validate', True),
            ('queue_verification', 'Name of the default queue for verification', 'str', 'tesla_verification', True),
            ('queue_alerts', 'Name of the default queue for alerts', 'str', 'tesla_alerts', True),
            ('queue_reporting', 'Name of the default queue for reports', 'str', 'tesla_reporting', True),
            ('queues', 'List of queues a worker will listen to.', 'list', None, False),
        )),
        ('rabbitmq', 'RabbitMQ configuration (only used for RabbitMQ deployment)', (
            ('erlang_cookie', 'Erlang Cookie value used for RabbitMQ cluster', 'str', None, True),
            ('admin_user', 'Administrator user to authenticate with RabbitMQ', 'str', None, True),
            ('admin_password', 'Administrator password to authenticate with RabbitMQ', 'str', None, True),
            ('admin_port', 'Port for the administration api of RabbitMQ', 'int', 15672, True),
            ('port', 'Port for the amqp api of RabbitMQ', 'int', 5672, True),
            ('host', 'Host name of the RabbitMQ server', 'str', 'localhost', True),
        )),
        ('django', 'DJango configuration', (
            ('secret_key', 'DJango secret value', 'str', None, None, True),
            ('allowed_hosts', 'Allowed hosts', 'list', [], None, True),
        )),
        ('deployment', 'Deployment configuration', (
            ('orchestrator', 'Docker orchestrator', 'enum', 'swarm', ['swarm', ], True),
            ('services', 'Deploy external services', 'bool', False, None, True),
            ('lb', 'Load balancer', 'enum', 'traefik', ['traefik', ], True),
            ('image', 'Docker image used for deployment', 'str',
             'teslace/core', None, True),
            ('version', 'Docker image version used for deployment', 'str', 'latest', None, True),
            ('secrets_path', 'Folder where secrets will be mounted', 'str', '/run/secrets', None, True),
            ('data_path', 'Folder where volumes will be persisted', 'str', '/var/tesla', None, True),
            ('specialized_workers', 'If enabled, specialized workers will be deployed.', 'bool', True, True),
        )),
        ('moodle', 'Moodle configuration. Only required if Moodle is deployed by TeSLA', (
            ('deploy', 'Deploy Moodle', 'bool', False, None, True),
            ('name', 'Unique VLE name in TeSLA', 'str', 'default_moodle', True),
            ('admin_user', 'Administrator user for Moodle', 'str', 'moodle', True),
            ('admin_password', 'Administrator password for Moodle', 'str', None, True),
            ('admin_email', 'Administrator email for Moodle', 'str', None, True),
            ('db_host', 'Host for Moodle database', 'str', None, None, True),
            ('db_port', 'Port for Moodle database', 'int', 3306, None, True),
            ('db_name', 'Database name for Moodle database', 'str', 'moodle', None, True),
            ('db_user', 'User for Moodle database', 'str', 'moodle', None, True),
            ('db_password', 'Password for Moodle database', 'str', None, None, True),
            ('db_prefix', 'Prefix for Moodle tables in the database', 'str', 'mdl_', None, True),
            ('cron_interval', 'Moodle Cron interval', 'int', 15, None, True),
            ('full_name', 'Full name for Moodle instance', 'str', 'TeSLA CE Moodle', None, True),
            ('short_name', 'Short name for Moodle instance', 'str', 'TeSLA CE', None, True),
            ('summary', 'Summary for Moodle instance', 'str', 'TeSLA CE Moodle Instance', None, True),
        ))
    )

    def __init__(self):
        """ Default class constructor """

        #: Configuration values contained in this class
        self._config = dict()

    def set(self, key, value=None):
        """
            Set a configuration value

            :param key: Configuration key
            :type key: str
            :param value: Configuration value
            :type value: object
            :exception TeslaConfigException: If the key is not valid or value type is incorrect
        """
        parts = self._get_field_parts(key)
        if parts is None:
            raise TeslaConfigException('Invalid configuration key {}'.format(key))
        field_desc = self._get_config_field(parts[0], parts[1])
        if field_desc is None:
            raise TeslaConfigException('Invalid configuration key {}'.format(key))
        self._config[key] = self._convert(value, field_desc)

    def get(self, key, default=None):
        """
            Get a configuration value

            :param key: Configuration key
            :type key: str
            :param default: Default value in case the key does not exist in the configuration
            :type default: object
            :return: Value assigned to the key or the default value provided
            :rtype: str | int | list | bool
            :exception TeslaConfigException: If the key is not valid
        """
        if key.upper() in self._config:
            return self._config[key.upper()]

        parts = self._get_field_parts(key)
        if parts is None:
            raise TeslaConfigException('Invalid configuration key {}'.format(key))
        field_desc = self._get_config_field(parts[0], parts[1])

        value = field_desc[3]
        if value is not None:
            return value

        return default

    def update(self, config):
        """
            Update configuration object with given options

            :param config: Configuration options
        """
        if isinstance(config, configparser.ConfigParser):
            # Update from config parser object
            for section in list(config):
                prefix = section.upper() + '_'
                if section == 'DEFAULT':
                    prefix = ''
                for s_key in config[section]:
                    key = s_key.upper()
                    if not key.startswith(prefix):
                        key = prefix + key
                    value = config.get(section, s_key, raw=True)
                    self.set(key, value)
        else:
            raise NotImplementedError

    def _get_config_field(self, section, key):
        """
            Get the description of a given configuration key

            :param section: Configuration section
            :param key: Configuration key
            :return: Tuple with field description or None if field does not exist
            :rtype: tuple
        """
        for sec in self._sections:
            if sec[0].upper() == section.upper():
                for s_key in sec[2]:
                    if s_key[0].upper() == key.upper():
                        return s_key
        return None

    @classmethod
    def get_sections(cls):
        """
            Get configuration sections

            :return: Configuration sections with values
            :rtype: dict
        """
        return cls._sections

    @staticmethod
    def _convert(value, desc):
        """
            Convert given value to the type of the field

            :param value: Provided value
            :type value: str | object
            :param desc: Field description tuple
            :type desc: tuple
            :return: Converted value
            :rtype: object
            :exception TeslaConfigException: If the value type is incorrect
        """
        if value is None:
            return None
        if isinstance(value, str) and len(value) == 0:
            return None
        field_type = desc[2]
        type_opt = desc[4]
        if field_type == 'str':
            return str(value)
        if field_type == 'bool':
            if isinstance(value, bool):
                return value
            if isinstance(value, int):
                return value == 1
            if isinstance(value, str):
                return value.upper() in ['TRUE', 'YES', 'Y', '1']
            raise TeslaConfigException('Invalid value for configuration key {}. Expected {}'.format(
                desc[0], field_type))
        if field_type == 'int':
            try:
                return int(value)
            except:
                raise TeslaConfigException('Invalid value for configuration key {}. Expected {}'.format(
                    desc[0], field_type))
        if field_type == 'enum':
            if value not in type_opt:
                raise TeslaConfigException('Invalid value for configuration key {}. Valid values are {}'.format(
                    desc[0], type_opt))
            return value
        if field_type == 'list':
            if isinstance(value, list):
                return value
            if isinstance(value, str):
                return value.split(',')
            raise TeslaConfigException('Invalid value for configuration key {}. Expected {}'.format(
                desc[0], field_type))
        # Valid types are processed on previous lines
        raise TeslaConfigException('Invalid type {} in field description'.format(field_type))

    def _get_field_parts(self, key):
        """
            Get the section and field values form a given configuration key
            :param key: Configuration key
            :type key: str
            :return: Tuple with section and field name or None if it is not a valid key
            :rtype: tuple
        """
        key = key.upper()
        for sec in self._sections:
            prefix = sec[0].upper() + '_'
            if key.startswith(prefix):
                section = sec[0].upper()
                field = key.split(prefix)[1]
                for s_field in sec[2]:
                    if field == s_field[0].upper():
                        return section, field
                return None
        return None

    def is_valid_key(self, key):
        """
            Check if a configuration key is valid or not

            :param key: The configuration key to be checked
            :type key: str
            :return: True if the key is valid or False otherwise
            :rtype: bool
        """
        return self._get_field_parts(key) is not None

    def get_config(self):
        """
            Get all configuration values

            :return: Dictionary with configuration
            :rtype: dict
        """
        conf = {}

        for section in self._sections:
            conf[section[0]] = {}
            for key in section[2]:
                conf[section[0]][key[0]] = {
                    'value': self.get('{}_{}'.format(section[0], key[0]).upper()),
                    'description': key[1]
                }
        return conf

    def get_config_kv(self):
        """
            Get all configuration values as a key-value dictionary

            :return: Dictionary with configuration kv
            :rtype: dict
        """
        conf = {}

        for section in self._sections:
            for key in section[2]:
                full_key = '{}_{}'.format(section[0], key[0]).upper()
                conf[full_key] = self.get(full_key)
        return conf

    def write(self, file_handler):
        """
            Write configuration to a file

            :param file_handle: File handler where configuration will be written
            :type file_handle: file
        """
        for section in self.get_sections():
            # Write the section header
            self._write_section_header(section, file_handler)
            for key in section[2]:
                # Skip keys with export flag to False
                if len(key) > 5 and not key[5]:
                    continue
                # Write key
                self._write_key(section, key, file_handler)
            file_handler.write('\n')

    @staticmethod
    def _write_section_header(section, file_handler):
        """
            Write configuration section header to a file

            :param section: Section description tuple
            :type section: tuple
            :param file_handle: File handler where it will be written
            :type file_handle: file
        """
        file_handler.write('# {}\n'.format(section[1]))
        file_handler.write('[{}]\n'.format(section[0]))

    @staticmethod
    def _write_key_header(key, file_handler):
        """
            Write configuration key header to a file

            :param key: Key description tuple
            :type key: tuple
            :param file_handle: File handler where it will be written
            :type file_handle: file
        """
        type = key[2]
        value = key[3]
        if type == 'enum':
            type = ' | '.join(key[4])
        elif type == 'list' and value is not None:
            value = ','.join(value)
        file_handler.write('# {}\n'.format(key[1]))
        file_handler.write('# ({})'.format(type))
        if value is not None and len(str(value)) > 0:
            file_handler.write(' default: {}'.format(value))
        file_handler.write('\n')

    def _write_key(self, section, key, file_handler):
        """
            Write configuration key information to a file

            :param section: Section description tuple
            :type section: tuple
            :param key: Key description tuple
            :type key: tuple
            :param file_handle: File handler where it will be written
            :type file_handle: file
        """
        # Write the key header
        self._write_key_header(key, file_handler)

        # Get current value
        value = self.get('{}_{}'.format(section[0], key[0]).upper())
        if key[2] == 'list' and value is not None:
            value = ','.join(value)
        if value is not None:
            value = str(value)

        # Get the default value as string
        default_value = key[3]
        if key[2] == 'list' and default_value is not None:
            default_value = ','.join(default_value)
        if default_value is not None:
            default_value = str(default_value)

        # Store the key
        if default_value is None and value is None:
            file_handler.write('{}=\n'.format(key[0]))
        elif value is not None and value != default_value:
            file_handler.write('{}={}\n'.format(key[0], value))
        else:
            file_handler.write('# {}={}\n'.format(key[0], default_value))



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Learner API Verification tasks module """
import simplejson
from celery import group
from celery.exceptions import Reject
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.db.models import Max
from django.db.models import Min

from tesla_ce import celery_app
from tesla_ce import models
from ..reports import update_learner_activity_instrument_report


@celery_app.task(ignore_result=True)
def create_request(activity_id, learner_id, path, instruments, session_id=None):
    try:
        # Store the request
        learner = models.Learner.objects.get(learner_id=learner_id)

        # Check Informed Consent status
        if not learner.ic_status.startswith('VALID'):
            default_storage.delete(path)
            raise Reject("Missing informed consent", requeue=False)

        req = models.Request.objects.create(learner_id=learner.id,
                                            activity_id=activity_id,
                                            data=path,
                                            session_id=session_id)
        # Add the instruments
        req.instruments.set(instruments)
        if len(instruments) != req.instruments.count():
            error = {
                "message": 'Invalid instruments',
            }
            default_storage.save('{}.error'.format(path), ContentFile(simplejson.dumps(error).encode('utf-8')))
            req.status = 2
            req.error_message = error["message"]
            req.save()
            raise Reject("Invalid Instruments", requeue=False)

        # Analyse this request
        verify_request.apply_async((req.id, ))
        default_storage.save('{}.valid'.format(path), ContentFile(simplejson.dumps({
            "message": "stored",
            "request_id": req.id
        }).encode('utf-8')))
        return req.id
    except models.Learner.DoesNotExist:
        raise Reject("Invalid Learner", requeue=False)


@celery_app.task(ignore_result=True)
def verify_request(request_id, result_id=None):
    """
        Start verification of a request

        :param request_id: Sample ID
        :type request_id: int
        :param result_id: Verification Request ID. Used by provider workers
        :type result_id: int
    """
    try:
        # Get the request
        request = models.Request.objects.get(id=request_id)

        # Create verification results for active providers
        request_task_group = []
        for instrument in request.instruments.all():
            instrument_task_group = []

            # Create the result for this instrument
            inst_result = models.RequestResult.objects.create(
                request_id=request_id,
                instrument_id=instrument.id
            )
            providers = models.Provider.objects.filter(instrument_id=instrument.id, enabled=True).all()
            if len(providers) == 0:
                inst_result.status = 4
                inst_result.save()
            else:
                for provider in providers:
                    req_result = models.RequestProviderResult.objects.create(
                        request=request,
                        provider=provider
                    )

                    try:
                        enrolment = models.Enrolment.objects.get(learner_id=request.learner_id, provider_id=provider.id)
                    except models.Enrolment.DoesNotExist:
                        enrolment = None

                    # check if instrument needs enrolment
                    if instrument.requires_enrolment is True and (enrolment is None or not enrolment.can_analyse):
                        req_result.status = 5
                        req_result.save()
                    else:
                        instrument_task_group.append(verify_request.s(request_id,
                                                                      req_result.id).set(queue=provider.queue))
                if len(instrument_task_group) == 0:
                    inst_result.status = 5
                    inst_result.save()
                else:
                    # Add instrument tasks to the request tasks
                    request_task_group += instrument_task_group

        if len(request_task_group) == 0:
            # There are no providers for any of the instruments
            if request.requestresult_set.filter(status=4).count() > 0:
                # Missing Provider
                request.status = 4
            elif request.requestresult_set.filter(status=5).count() > 0:
                # Missing Enrolment
                request.status = 5
            else:
                # Error
                request.status = 2
            request.save()
        else:
            # Launch processing tasks
            group(request_task_group).apply_async()

    except models.Request.DoesNotExist:
        raise Reject("Invalid Request", requeue=False)


@celery_app.task(ignore_result=True, bind=True)
def create_verification_summary(self, request_id, instrument_id):
    """
        Summarize request verification
        :param request_id: Request ID
        :type request_id: int
        :param request_id: Instrument ID
        :type request_id: int
    """
    try:
        result = models.RequestResult.objects.get(request_id=request_id, instrument_id=instrument_id)
    except models.RequestResult.DoesNotExist:
        result = models.RequestResult.objects.create(request_id=request_id, instrument_id=instrument_id,
                                                     code=0, status=0)

    res_values = models.RequestProviderResult.objects.filter(request_id=request_id,
                                                             provider__instrument_id=instrument_id,
                                                             status=1
                                                             ).aggregate(Max('result'), Max('code'))
    status_values = models.RequestProviderResult.objects.filter(request_id=request_id,
                                                                provider__instrument_id=instrument_id
                                                               ).aggregate(Min('status'), Max('status'))

    # Update status. If here are pending requests, set as PENDING, otherwise to the maximum status level
    result.status = 0
    if status_values['status__min'] > 0:
        result.status = status_values['status__max']
    # Compute average result for providers
    result.result = res_values['result__max']
    # Set the maximum alert level as the final level
    result.code = res_values['code__max']
    result.save()

    # If there are no missing requests for this learner and instrument in the activity, update the activity report
    if result.status > 0:
        # Check if all the instruments finished
        global_status_values = models.RequestProviderResult.objects.filter(
            request_id=request_id
        ).aggregate(Min('status'), Max('status'))

        # Update status. If there are pending requests, set as PENDING, otherwise to the maximum status level
        if global_status_values['status__min'] == 0:
            # There are pending results
            if global_status_values['status__max'] > 0:
                result.request.status = 2  # Processing
            else:
                result.request.status = 1  # Scheduled
        else:
            if global_status_values['status__max'] < 5:
                # Processed, Error, Timeout or Missing Provider
                result.request.status = global_status_values['status__max'] + 2
            else:
                result.request.status = 4  # Error
        result.request.save()

        # Update reports
        update_learner_activity_instrument_report.apply_async((result.request.learner_id,
                                                               result.request.activity_id,
                                                               instrument_id,))



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
import traceback

from django.conf import settings
from django.http import HttpResponse

from .lib.exception import tesla_report_exception


class TeslaErrorHandlerMiddleware:
    """
        Error handles middleware
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        return response

    def process_exception(self, request, exception):
        if not settings.DEBUG:
            if exception:
                # Format your message here
                message = "**{url}**\n\n{error}\n\n````{tb}````".format(
                    url=request.build_absolute_uri(),
                    error=repr(exception),
                    tb=traceback.format_exc()
                )
                # Do now whatever with this message
                # e.g. requests.post(<slack channel/teams channel>, data=message)
                print(message)
                tesla_report_exception(exception)

            return HttpResponse("Error processing the request.", status=500)



# Generated by Django 3.2.4 on 2021-10-09 00:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0019_auto_20210907_1602'),
    ]

    operations = [
        migrations.AddField(
            model_name='institution',
            name='allowed_domains',
            field=models.TextField(blank=True, default=None, help_text='Allowed domains for redirection', null=True),
        ),
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Learner verification samples views module """
from django.utils import timezone

from tesla_ce import tasks
from .base import BaseLearnerAPIView
from ..serializers.verification import VerificationSampleSerializer


class LearnerVerification(BaseLearnerAPIView):
    """
        Provide a verification request
    """
    serializer_class = VerificationSampleSerializer

    def get_storing_path(self, institution_id, data):
        """
            Create the path where this request should be stored
            :param institution_id: Id of the institution
            :type institution_id: int
            :param data: The data of the request
            :type data: dict
            :return: Path to the storage
            :rtype: str
        """
        prefix = timezone.now().strftime("%Y%m%d_%H%M%S_%f")
        metadata = data.get('metadata')
        if metadata is None:
            filename = 'request'
        else:
            filename = metadata.get('filename', 'document')
        if data.get('session_id') is None:
            # Attachment
            path = '{}/requests/{}/{}/{}/documents/{}_{}'.format(
                institution_id,
                data['learner_id'],
                data['course_id'],
                data['activity_id'],
                prefix,
                filename)
        else:
            # Sample
            path = '{}/requests/{}/{}/{}/sessions/{}/{}_{}'.format(
                institution_id,
                data['learner_id'],
                data['course_id'],
                data['activity_id'],
                data['session_id'],
                prefix,
                filename
            )
        return path

    def enqueue_request(self, data, path):
        """
            Enqueue a new task to manage new storage data
            :param data: Stored data
            :type data: dict
            :param path: Path where the data has been stored
            :type path: str
        """
        kwargs = {
            'activity_id': data['activity_id'],
            'learner_id': data['learner_id'],
            'path': path,
            'instruments': data['instruments'],
            'session_id': data.get('session_id')
        }
        tasks.requests.verification.create_request.apply_async(kwargs=kwargs)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" DJango command to generate the configuration file for TeSLA system """
from ..base import TeslaDeployCommand


class Command(TeslaDeployCommand):
    """ Command to generate deployment files """
    help = 'Generates configuration files to deploy the TeSLA CE system core'

    def custom_handle(self):
        """
            Custom actions for this command
        """
        # Export the deployment scripts
        self.client.export_deployment_scripts(output=self._options['out'], mode=self._options['mode'])
        self.stdout.write(self.style.SUCCESS('Deployment scripts written at {}'.format(self._options['out'])))



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" DJango settings module """
import os

from configurations import Configuration

from tesla_ce.lib import ConfigManager
from tesla_ce.lib import VaultManager
from tesla_ce.lib.checks import check_vault_connection
from tesla_ce.lib.exception import TeslaConfigException

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'tesla_ce.settings')
os.environ.setdefault('DJANGO_CONFIGURATION', 'Production')


class BaseConfiguration(Configuration):
    """ Base configuration transversal to all options """
    # Build paths inside the project like this: os.path.join(BASE_DIR, ...)
    BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    # Allowed hosts to access DJango application
    ALLOWED_HOSTS = []

    # Disable Prometheus migrations
    PROMETHEUS_EXPORT_MIGRATIONS = False

    # Application definition
    INSTALLED_APPS = [
        'django.contrib.admin',
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'django.contrib.sessions',
        'django.contrib.messages',
        'django.contrib.staticfiles',
        # Celery results
        'django_celery_results',
        'django_celery_beat',
        # Include DJango rest framework
        'rest_framework',
        'django_filters',
        # CORS framework
        'corsheaders',
        # API documentation
        'drf_yasg',
        # Prometheus metrics exporter
        'django_prometheus',
        # Include TeSLA CE app
        'tesla_ce',
    ]

    MIDDLEWARE = [
        'django_prometheus.middleware.PrometheusBeforeMiddleware',
        'corsheaders.middleware.CorsMiddleware',
        'django.middleware.security.SecurityMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.common.CommonMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'corsheaders.middleware.CorsPostCsrfMiddleware',
        #'tesla_ce.lib.auth.tesla_token_auth_middleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
        'tesla_ce.error.TeslaErrorHandlerMiddleware',
        'django_prometheus.middleware.PrometheusAfterMiddleware',
    ]

    TEMPLATES = [
        {
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'DIRS': [],
            'APP_DIRS': True,
            'OPTIONS': {
                'context_processors': [
                    'django.template.context_processors.debug',
                    'django.template.context_processors.request',
                    'django.contrib.auth.context_processors.auth',
                    'django.contrib.messages.context_processors.messages',
                ],
            },
        },
    ]

    WSGI_APPLICATION = 'tesla_ce.wsgi.application'
    ROOT_URLCONF = 'tesla_ce.urls'

    # Static files (CSS, JavaScript, Images)
    # https://docs.djangoproject.com/en/2.2/howto/static-files/
    PROJECT_DIR = os.path.dirname(os.path.abspath(__file__))
    STATIC_ROOT = os.path.join(PROJECT_DIR, 'static')
    STATIC_URL = '/static/'

    # Internationalization
    # https://docs.djangoproject.com/en/2.2/topics/i18n/
    LANGUAGE_CODE = 'en-us'
    TIME_ZONE = 'UTC'
    USE_I18N = True
    USE_L10N = True
    USE_TZ = True

    # If your Django app is behind a proxy that sets a header to specify secure
    # connections, AND that proxy ensures that user-submitted headers with the
    # same name are ignored (so that people can't spoof it), set this value to
    # a tuple of (header_name, header_value). For any requests that come in with
    # that header/value, request.is_secure() will return True.
    # WARNING! Only set this if you fully understand what you're doing. Otherwise,
    # you may be opening yourself up to a security risk.
    SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')

    # Celery configuration
    CELERY_RESULT_BACKEND = 'django-db'
    CELERY_CACHE_BACKEND = 'django-cache'
    CELERY_BEAT_SCHEDULER = 'django_celery_beat.schedulers:DatabaseDatabaseScheduler'

    # Enabled TeSLA modules
    TESLA_MODULES = []

    REST_FRAMEWORK = {
        'DEFAULT_AUTHENTICATION_CLASSES': (
            'tesla_ce.lib.auth.JWTAuthentication',
        ),
        'DEFAULT_RENDERER_CLASSES': (
            'rest_framework.renderers.JSONRenderer',
            'rest_framework.renderers.BrowsableAPIRenderer',
            'rest_framework_datatables.renderers.DatatablesRenderer',
        ),
        'DEFAULT_VERSIONING_CLASS': 'rest_framework.versioning.NamespaceVersioning',
        'DEFAULT_VERSION': 'v2',
        'ALLOWED_VERSIONS': ['v1', 'v2'],
        'DEFAULT_PAGINATION_CLASS': 'rest_framework_datatables.pagination.DatatablesLimitOffsetPagination',
        'PAGE_SIZE': 10
    }

    # TeSLA Password checker
    # DJANGO: User default DJANGO engine
    # VAULT: User vault credentials manager
    #
    TESLA_PASSWORD_BACKEND = os.getenv('TESLA_PASSWORD_BACKEND', 'DJANGO').upper()

    # TeSLA Authentication configuration
    TESLA_JWT_TOKEN = 'JWT'
    CORS_ORIGIN_ALLOW_ALL = True
    CORS_ALLOW_CREDENTIALS = False

    # todo: review this headers when V1 is deprecated
    CORS_ALLOW_HEADERS = (
        'vleid',
        'VLEID',
        'Authorization',
        'Content-Type'
    )


class Production(BaseConfiguration):
    """
        Production configuration class.
    """
    #: TeSLA configuration
    TESLA_CONFIG = ConfigManager()

    @classmethod
    def _add_storage(cls):
        # Your Amazon Web Services access key, as a string.
        cls.AWS_ACCESS_KEY_ID = cls.TESLA_CONFIG.config.get('STORAGE_ACCESS_KEY')
        # Your Amazon Web Services secret access key, as a string.
        cls.AWS_SECRET_ACCESS_KEY = cls.TESLA_CONFIG.config.get('STORAGE_SECRET_KEY')
        # Your Amazon Web Services storage bucket name, as a string.
        cls.AWS_STORAGE_BUCKET_NAME = cls.TESLA_CONFIG.config.get('STORAGE_BUCKET_NAME')
        # (optional, None or canned ACL, default public-read) Must be either None or from the list of canned ACLs.
        # If set to None then all files will inherit the bucket’s ACL.
        cls.AWS_DEFAULT_ACL = 'private'
        # (optional: default is None) Name of the AWS S3 region to use (eg. eu-west-1)
        cls.AWS_S3_REGION_NAME = cls.TESLA_CONFIG.config.get('STORAGE_REGION')

        # Custom S3 URL to use when connecting to S3, including scheme. Overrides AWS_S3_REGION_NAME and AWS_S3_USE_SSL.
        cls.AWS_S3_ENDPOINT_URL = cls.TESLA_CONFIG.config.get('STORAGE_URL')

        # Disable SSL verification
        if cls.TESLA_CONFIG.config.get('STORAGE_SSL_VERIFY') is False:
            cls.AWS_S3_VERIFY = False

        # Get the public bucket name
        cls.AWS_S3_STORAGE_PUBLIC_BUCKET_NAME = cls.TESLA_CONFIG.config.get('STORAGE_PUBLIC_BUCKET_NAME')

        if cls.AWS_ACCESS_KEY_ID is not None and cls.AWS_SECRET_ACCESS_KEY is not None:
            cls.DEFAULT_FILE_STORAGE = 'storages.backends.s3boto3.S3Boto3Storage'
            cls.STATICFILES_STORAGE = 'tesla_ce.lib.storage.bucket.PublicS3Boto3Storage'
            cls.STATIC_ROOT = '/'
        else:
            cls.DEFAULT_FILE_STORAGE = 'django.core.files.storage.FileSystemStorage'

    @classmethod
    def pre_setup(cls):
        # Load DJango default configuration
        super().pre_setup()

        if len(cls.TESLA_CONFIG.modules) > 0:
            # Starting in production mode
            cls.TESLA_CONFIG.config.set('TESLA_MODE', 'production')

            # Enable modules
            cls.TESLA_MODULES = []
            for module in cls.TESLA_CONFIG.modules:
                cls.TESLA_MODULES.append(module)
                apps = cls.TESLA_CONFIG.modules[module].get('apps', [])
                for app in apps:
                    if app not in cls.INSTALLED_APPS:
                        cls.INSTALLED_APPS.append(app)
                cls.TESLA_CONFIG.config.set('CELERY_QUEUES', cls.TESLA_CONFIG.modules[module].get('queues'))

            # Apply configuration
            cls.SECRET_KEY = cls.TESLA_CONFIG.config.get('DJANGO_SECRET_KEY')
            cls.ALLOWED_HOSTS = cls.TESLA_CONFIG.config.get('DJANGO_ALLOWED_HOSTS')
            cls.DATABASES = cls.TESLA_CONFIG.get_database_config()
            cls.CACHES = cls.TESLA_CONFIG.get_cache_config()
            cls._add_storage()
            cls.CELERY_BROKER_URL = cls.TESLA_CONFIG.get_celery_config()
        else:
            # Starting in configuration mode
            cls.TESLA_CONFIG.config.set('TESLA_MODE', 'config')
            cls.TESLA_CONFIG.config.set('VAULT_MANAGEMENT', True)
            cls.SECRET_KEY = "Configuration Key"
            cls.DATABASES = {
                'default': {
                    #'ENGINE': 'django.db.backends.{}'.format(cls.TESLA_CONFIG.config.get('DB_ENGINE')),
                    'ENGINE': 'django_prometheus.db.backends.{}'.format(cls.TESLA_CONFIG.config.get('DB_ENGINE')),
                    'HOST': cls.TESLA_CONFIG.config.get('DB_HOST'),
                    'NAME': cls.TESLA_CONFIG.config.get('DB_NAME'),
                    'USER': cls.TESLA_CONFIG.config.get('DB_USER'),
                    'PASSWORD': cls.TESLA_CONFIG.config.get('DB_PASSWORD'),
                    'PORT': cls.TESLA_CONFIG.config.get('DB_PORT'),
                },
                'admin': {
                    #'ENGINE': 'django.db.backends.{}'.format(cls.TESLA_CONFIG.config.get('DB_ENGINE')),
                    'ENGINE': 'django_prometheus.db.backends.{}'.format(cls.TESLA_CONFIG.config.get('DB_ENGINE')),
                    'HOST': cls.TESLA_CONFIG.config.get('DB_HOST'),
                    'NAME': cls.TESLA_CONFIG.config.get('DB_NAME'),
                    'USER': cls.TESLA_CONFIG.config.get('DB_ROOT_USER'),
                    'PASSWORD': cls.TESLA_CONFIG.config.get('DB_ROOT_PASSWORD'),
                    'PORT': cls.TESLA_CONFIG.config.get('DB_PORT'),
                }
            }
            cls.CACHES = cls.TESLA_CONFIG.get_cache_config()
            cls._add_storage()
            cls.CELERY_BROKER_URL = cls.TESLA_CONFIG.get_celery_config()

        # Add the URLs to the API and DASHBOARD
        cls.API_URL = 'https://{}'.format(cls.TESLA_CONFIG.config.get('TESLA_DOMAIN'))
        cls.DASHBOARD_URL = 'https://{}'.format(cls.TESLA_CONFIG.config.get('TESLA_DOMAIN'))

        # If no database is selected, remove celery results backend and change to scope auth middleware
        if len(cls.DATABASES) == 0:
            cls.CELERY_RESULT_BACKEND = None

        # Use debug API authentication.
        cls.DEBUG_AUTHENTICATION = False

        # Use debug API authentication object.
        # Can be a pair <type,id> to set desired authentication object
        # Or None for unauthenticated access
        cls.DEBUG_AUTHENTICATION_OBJECT = None


class Dev(Production):
    """
        Development configuration.
    """
    @classmethod
    def pre_setup(cls):
        # Check if Vault is available
        vault_connection_report = check_vault_connection(config=cls.TESLA_CONFIG.config)

        if os.getenv('SSLSERVER') in ['True', 'true', 1, '1']:
            cls.INSTALLED_APPS.append('sslserver')

        # If vault is enabled, apply development actions
        if vault_connection_report['connected']:
            # Unseal Vault if it is sealed
            if vault_connection_report['initialized'] and vault_connection_report['sealed']:
                keys = cls.TESLA_CONFIG.config.get('VAULT_KEYS')
                if keys is not None:
                    cls.TESLA_CONFIG.config.set('VAULT_MANAGEMENT', True)
                    manager = VaultManager(config=cls.TESLA_CONFIG)
                    manager.unseal()
                    # Update vault status
                    vault_connection_report = check_vault_connection(config=cls.TESLA_CONFIG.config)
            # Allow to start as any module
            force_module = os.getenv('TESLA_RUN_AS_MODULE', None)
            if force_module is not None:
                token = cls.TESLA_CONFIG.config.get('VAULT_TOKEN')

                # Check that Vault is ready and token is provided
                if not vault_connection_report['ready']:
                    raise TeslaConfigException('Requested to start as module {}, but Vault is not ready'.format(
                        force_module))
                if token is None:
                    raise TeslaConfigException('Requested to start as module {}, but Vault token is missing'.format(
                        force_module))
                cls.TESLA_CONFIG.config.set('VAULT_MANAGEMENT', True)
                manager = VaultManager(config=cls.TESLA_CONFIG)
                modules = force_module.split(',')
                role_id = []
                secret_id = []
                for module in modules:
                    creds = manager.get_module_credentials(module)
                    role_id.append(creds['role_id'])
                    secret_id.append(creds['secret_id'])
                cls.TESLA_CONFIG.load_vault(role_id=role_id, secret_id=secret_id)

        # Load Production configuration
        super().pre_setup()

        if os.getenv('API_URL') is not None:
            cls.API_URL = os.getenv('API_URL')
        if os.getenv('DASHBOARD_URL') is not None:
            cls.DASHBOARD_URL = os.getenv('DASHBOARD_URL')

        # Enable Debug
        cls.DEBUG = True

        # Enable all hosts
        cls.ALLOWED_HOSTS = ['*']

        # Use debug API authentication.
        cls.DEBUG_AUTHENTICATION = os.getenv('DEBUG_AUTHENTICATION', False) in [1, "1", True, "True", "true"]

        if cls.DEBUG_AUTHENTICATION:
            cls.REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES'] = ('tesla_ce.lib.auth.DebugJWTAuthentication',)

        # Use debug API authentication object.
        # Can be a pair <type,id> to set desired authentication object. Type must be: admin, user, vle, provider
        # Or None for unauthenticated access
        cls.DEBUG_AUTHENTICATION_OBJECT = os.getenv('DEBUG_AUTHENTICATION_OBJECT', None)
        if cls.DEBUG_AUTHENTICATION_OBJECT is not None:
            cls.DEBUG_AUTHENTICATION_OBJECT = cls.DEBUG_AUTHENTICATION_OBJECT.replace(" ", "").split(",")
            if cls.DEBUG_AUTHENTICATION_OBJECT[0] not in ["admin", "user", "vle", "provider"]:
                raise Exception(
                    "Invalid authentication object {}. Allowed values are: admin, user, vle, provider".format(
                        cls.DEBUG_AUTHENTICATION_OBJECT[0]
                    )
                )


class Test(Production):
    """
        Test configuration.
    """
    @classmethod
    def pre_setup(cls):
        # Load Production configuration
        super().pre_setup()

        if 'tesla_ce.apps.api' not in cls.INSTALLED_APPS:
            cls.INSTALLED_APPS += ['tesla_ce.apps.api']
        if 'tesla_ce.apps.lapi' not in cls.INSTALLED_APPS:
            cls.INSTALLED_APPS += ['tesla_ce.apps.lapi']

        if 'api' not in cls.TESLA_MODULES:
            cls.TESLA_MODULES += ['api']
        if 'lapi' not in cls.TESLA_MODULES:
            cls.TESLA_MODULES += ['lapi']



# Generated by Django 3.2.4 on 2021-06-16 18:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0014_uioption'),
    ]

    operations = [
        migrations.AddField(
            model_name='uioption',
            name='roles',
            field=models.CharField(blank=True, help_text='Required roles.', max_length=250, null=True),
        ),
    ]
